@@ -1,7 +1,9 @@
+from dataclasses import replace
 from io import BytesIO
 import json
 import webbrowser
 from pathlib import Path
+from configparser import ParsingError
 
 from databricks.labs.blueprint.cli import App
 from databricks.labs.blueprint.entrypoint import get_logger
@@ -17,8 +19,7 @@ from databricks.labs.ucx.contexts.account_cli import AccountContext
 from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext, LocalCheckoutContext
 from databricks.labs.ucx.hive_metastore.tables import What
 from databricks.labs.ucx.install import AccountInstaller
-from databricks.labs.ucx.source_code.linters.files import LocalCodeLinter
-
+from databricks.labs.ucx.workspace_access.groups import AccountGroupLookup
 
 ucx = App(__file__)
 logger = get_logger(__file__)
@@ -34,7 +35,11 @@ def _get_workspace_contexts(
 ) -> list[WorkspaceContext]:
     """Get workspace contexts to the workspaces for which the user has access"""
     if not a:
-        a = AccountClient(product='ucx', product_version=__version__)
+        try:
+            a = AccountClient(product='ucx', product_version=__version__)
+        except ParsingError as e:
+            logger.error("Could not create account client", exc_info=e)
+            return []
     account_installer = AccountInstaller(a)
     workspace_contexts = account_installer.get_workspace_contexts(w, run_as_collection, **named_parameters)
     return workspace_contexts
@@ -164,7 +169,7 @@ def create_account_groups(
     """
     if not ctx:
         ctx = AccountContext(a, named_parameters)
-    ctx.account_workspaces.create_account_level_groups(prompts, ctx.workspace_ids)
+    ctx.account_workspaces.create_account_level_groups(prompts)
 
 
 @ucx.command
@@ -250,7 +255,8 @@ def update_migration_progress(
 def repair_run(w: WorkspaceClient, step):
     """Repair Run the Failed Job"""
     if not step:
-        raise KeyError("You did not specify --step")
+        logger.error("--step is a required parameter")
+        return
     ctx = WorkspaceContext(w)
     logger.info(f"Repair Running {step} Job")
     ctx.deployed_workflows.repair_run(step)
@@ -308,10 +314,12 @@ def move(
     """move a uc table/tables from one schema to another schema in same or different catalog"""
     logger.info("Running move command")
     if from_catalog == "" or to_catalog == "":
-        logger.error("Please enter from_catalog and to_catalog details")
+        logger.error("--from_catalog and --to_catalog are required parameter")
         return
     if from_schema == "" or to_schema == "" or from_table == "":
-        logger.error("Please enter from_schema, to_schema and from_table (enter * for migrating all tables) details.")
+        logger.error(
+            "--from_schema, --to_schema and --from_table (enter * for migrating all tables) are required parameter."
+        )
         return
     if from_catalog == to_catalog and from_schema == to_schema:
         logger.error("please select a different schema or catalog to migrate to")
@@ -337,10 +345,12 @@ def alias(
 ):
     """move a uc table/tables from one schema to another schema in same or different catalog"""
     if from_catalog == "" or to_catalog == "":
-        logger.error("Please enter from_catalog and to_catalog details")
+        logger.error("--from_catalog and --to_catalog are required parameter")
         return
     if from_schema == "" or to_schema == "" or from_table == "":
-        logger.error("Please enter from_schema, to_schema and from_table (enter * for migrating all tables) details.")
+        logger.error(
+            "--from_schema, --to_schema and --from_table (enter * for migrating all tables) are required parameter."
+        )
         return
     if from_catalog == to_catalog and from_schema == to_schema:
         logger.error("please select a different schema or catalog to migrate to")
@@ -495,6 +505,27 @@ def migrate_credentials(
         raise ValueError("Unsupported cloud provider")
 
 
+@ucx.command
+def migrate_glue_credentials(
+    w: WorkspaceClient,
+    prompts: Prompts,
+    ctx: WorkspaceContext | None = None,
+    run_as_collection: bool = False,
+    a: AccountClient | None = None,
+    **named_parameters,
+):
+    """For AWS, this command migrates AWS Glue IAM roles to UC storage credentials. The AWS Glue IAM roles to location"""
+    workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
+    if ctx:
+        workspace_contexts = [ctx]
+    if w.config.is_aws:
+        for workspace_ctx in workspace_contexts:
+            logger.info(f"Running cmd for workspace {workspace_ctx.workspace_client.get_workspace_id()}")
+            workspace_ctx.iam_role_migration.migrate_glue(prompts)
+    else:
+        raise ValueError("Unsupported cloud provider")
+
+
 @ucx.command
 def migrate_locations(
     w: WorkspaceClient,
@@ -578,16 +609,6 @@ def revert_cluster_remap(w: WorkspaceClient, prompts: Prompts):
     ctx.cluster_access.revert_cluster_remap(cluster_list, cluster_ids)
 
 
-@ucx.command
-def migrate_local_code(w: WorkspaceClient, prompts: Prompts):
-    """Fix the code files based on their language."""
-    ctx = LocalCheckoutContext(w)
-    working_directory = Path.cwd()
-    if not prompts.confirm("Do you want to apply UC migration to all files in the current directory?"):
-        return
-    ctx.local_file_migrator.apply(working_directory)
-
-
 @ucx.command(is_account=True)
 def show_all_metastores(a: AccountClient, workspace_id: str | None = None):
     """Show all metastores in the account"""
@@ -635,6 +656,32 @@ def create_ucx_catalog(w: WorkspaceClient, prompts: Prompts, ctx: WorkspaceConte
     workspace_context.verify_progress_tracking.verify()
 
 
+@ucx.command
+def assign_owner_group(
+    w: WorkspaceClient,
+    prompts: Prompts,
+    *,
+    ctx: WorkspaceContext | None = None,
+    run_as_collection: bool = False,
+    a: AccountClient | None = None,
+) -> None:
+    """
+    Pick owner group. This group will be assigned as owner to all the migrated tables.
+    """
+    if ctx:
+        workspace_contexts = [ctx]
+    else:
+        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
+
+    owner_group = AccountGroupLookup(workspace_contexts[0].workspace_client).pick_owner_group(prompts)
+    if not owner_group:
+        return
+    for workspace_context in workspace_contexts:
+        config = workspace_context.installation.load(WorkspaceConfig)
+        config = replace(config, default_owner_group=owner_group)
+        workspace_context.installation.save(config)
+
+
 @ucx.command
 def migrate_tables(
     w: WorkspaceClient,
@@ -655,16 +702,7 @@ def migrate_tables(
     for workspace_context in workspace_contexts:
         deployed_workflows = workspace_context.deployed_workflows
 
-        try:
-            workspace_context.verify_progress_tracking.verify()
-        except RuntimeWarning:
-            logger.warning(
-                "We couldn't detect a successful run of the assessment workflow."
-                "The assessment workflow is a prerequisite for the migrate-tables workflow."
-                "It can be run by using the `ensure-assessment-run` command."
-            )
-            return
-
+        workspace_context.verify_progress_tracking.verify()
         deployed_workflows.run_workflow("migrate-tables")
 
         tables = list(workspace_context.tables_crawler.snapshot())
@@ -705,7 +743,26 @@ def migrate_acls(
         workspace_contexts = [ctx]
     else:
         workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
-    target_catalog, hms_fed = named_parameters.get("target_catalog"), named_parameters.get("hms_fed", False)
+    target_catalog = named_parameters.get("target_catalog")
+    hms_fed = named_parameters.get("hms_fed", False)
+    dry_run = named_parameters.get("dry_run", False)
+    if dry_run:
+        total_grants = 0
+        for workspace_context in workspace_contexts:
+            grants = workspace_context.acl_migrator.snapshot()
+            total_grants += len(grants)
+        logger.info(
+            f"Dry run completed. Found {total_grants} grants. The crawled grants can be found in the 'inferred_grants' table. "
+            "No changes were made."
+        )
+        urls: str = ""
+        for workspace_context in workspace_contexts:
+            urls += (
+                f"{workspace_context.connect_config.host}/explore/data/hive_metastore/"
+                f"{workspace_context.config.inventory_database}/inferred_grants\n"
+            )
+        logger.info(f"URLs to the inferred grants tables: \n{urls}")
+        return
     for workspace_context in workspace_contexts:
         workspace_context.acl_migrator.migrate_acls(target_catalog=target_catalog, hms_fed=hms_fed)
 
@@ -724,14 +781,21 @@ def migrate_dbsql_dashboards(
     else:
         workspace_contexts = _get_workspace_contexts(w, a, run_as_collection)
     for workspace_context in workspace_contexts:
-        workspace_context.redash.migrate_dashboards(dashboard_id)
+        if dashboard_id:
+            workspace_context.redash.migrate_dashboards(dashboard_id)
+        else:
+            workspace_context.redash.migrate_dashboards()
 
 
 @ucx.command
-def revert_dbsql_dashboards(w: WorkspaceClient, dashboard_id: str | None = None):
+def revert_dbsql_dashboards(w: WorkspaceClient, dashboard_id: str | None = None, ctx: WorkspaceContext | None = None):
     """Revert migrated DBSQL Dashboard queries back to their original state"""
-    ctx = WorkspaceContext(w)
-    ctx.redash.revert_dashboards(dashboard_id)
+    ctx = ctx or WorkspaceContext(w)
+    ctx.redash_crawler.snapshot(force_refresh=True)  # Need the latest tags before reverting dashboards
+    if dashboard_id:
+        ctx.redash.revert_dashboards(dashboard_id)
+    else:
+        ctx.redash.revert_dashboards()
 
 
 @ucx.command(is_account=True)
@@ -803,13 +867,43 @@ def download(
 
 @ucx.command
 def lint_local_code(
-    w: WorkspaceClient, prompts: Prompts, path: str | None = None, ctx: LocalCheckoutContext | None = None
+    w: WorkspaceClient,
+    prompts: Prompts,
+    path: Path | str | None = None,
+    ctx: LocalCheckoutContext | None = None,
 ):
     """Lint local code files looking for problems."""
     if ctx is None:
         ctx = LocalCheckoutContext(w)
-    linter: LocalCodeLinter = ctx.local_code_linter
-    linter.lint(prompts, None if path is None else Path(path))
+    if not path:
+        response = prompts.question(
+            "Which file or directory do you want to lint?",
+            default=Path.cwd().as_posix(),
+            validate=lambda p_: Path(p_).exists(),
+        )
+        assert response
+        path = Path(response)
+    for advice in ctx.local_code_linter.lint(Path(path)):
+        print(advice)
+
+
+@ucx.command
+def migrate_local_code(
+    w: WorkspaceClient, prompts: Prompts, path: Path | str | None = None, ctx: LocalCheckoutContext | None = None
+):
+    """Fix the code files based on their language."""
+    if ctx is None:
+        ctx = LocalCheckoutContext(w)
+    if not path:
+        response = prompts.question(
+            "Which file or directory do you want to lint?",
+            default=Path.cwd().as_posix(),
+            validate=lambda p_: Path(p_).exists(),
+        )
+        assert response
+        path = Path(response)
+    for advice in ctx.local_code_linter.apply(Path(path)):
+        print(advice)
 
 
 @ucx.command
@@ -821,10 +915,10 @@ def export_assessment(w: WorkspaceClient, prompts: Prompts):
 
 
 @ucx.command
-def create_federated_catalog(w: WorkspaceClient, _: Prompts):
+def create_federated_catalog(w: WorkspaceClient, prompts: Prompts):
     """(Experimental) Create federated catalog from current workspace Hive Metastore."""
     ctx = WorkspaceContext(w)
-    ctx.federation.register_internal_hms_as_federated_catalog()
+    ctx.federation.create_from_cli(prompts)
 
 
 @ucx.command
@@ -835,5 +929,28 @@ def enable_hms_federation(w: WorkspaceClient, _: Prompts, ctx: WorkspaceContext
     ctx.federation_enabler.enable()
 
 
+@ucx.command
+def migrate_dlt_pipelines(
+    w: WorkspaceClient,
+    ctx: WorkspaceContext | None = None,
+    run_as_collection: bool = False,
+    a: AccountClient | None = None,
+    **named_parameters,
+) -> None:
+    """
+    Migrate DLT pipelines to UC that were crawled in the assessment workflow
+    Flags
+    -- include-pipeline-ids: Comma separated list of pipeline ids to migrate
+    -- exclude-pipeline-ids: Comma separated list of pipeline ids to exclude from migration
+    """
+    if ctx:
+        workspace_contexts = [ctx]
+    else:
+        workspace_contexts = _get_workspace_contexts(w, a, run_as_collection, **named_parameters)
+
+    for workspace_context in workspace_contexts:
+        workspace_context.pipelines_migrator.migrate_pipelines()
+
+
 if __name__ == "__main__":
     ucx()
