@@ -12,12 +12,16 @@ from databricks.labs.blueprint.tui import Prompts
 from databricks.labs.blueprint.wheels import ProductInfo, WheelsV2
 from databricks.labs.lsql.backends import SqlBackend
 
+from databricks.labs.ucx.assessment.dashboards import DashboardOwnership
+from databricks.labs.ucx.assessment.jobs import JobsCrawler
+from databricks.labs.ucx.assessment.pipelines import PipelinesCrawler
+from databricks.labs.ucx.hive_metastore.pipelines_migrate import PipelinesMigrator
 from databricks.labs.ucx.recon.data_comparator import StandardDataComparator
 from databricks.labs.ucx.recon.data_profiler import StandardDataProfiler
 from databricks.labs.ucx.recon.metadata_retriever import DatabricksTableMetadataRetriever
 from databricks.labs.ucx.recon.migration_recon import MigrationRecon
 from databricks.labs.ucx.recon.schema_comparator import StandardSchemaComparator
-from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler
+from databricks.labs.ucx.source_code.directfs_access import DirectFsAccessCrawler, DirectFsAccessOwnership
 from databricks.labs.ucx.source_code.python_libraries import PythonLibraryResolver
 from databricks.labs.ucx.source_code.used_table import UsedTablesCrawler
 from databricks.sdk import AccountClient, WorkspaceClient, core
@@ -25,10 +29,11 @@ from databricks.sdk.service import sql
 
 from databricks.labs.ucx.account.workspaces import WorkspaceInfo
 from databricks.labs.ucx.assessment.azure import AzureServicePrincipalCrawler
+from databricks.labs.ucx.assessment.dashboards import LakeviewDashboardCrawler, RedashDashboardCrawler
 from databricks.labs.ucx.assessment.export import AssessmentExporter
 from databricks.labs.ucx.aws.credentials import CredentialManager
 from databricks.labs.ucx.config import WorkspaceConfig
-from databricks.labs.ucx.framework.owners import AdministratorLocator
+from databricks.labs.ucx.framework.owners import AdministratorLocator, WorkspacePathOwnership, LegacyQueryOwnership
 from databricks.labs.ucx.hive_metastore import ExternalLocations, MountsCrawler, TablesCrawler
 from databricks.labs.ucx.hive_metastore.catalog_schema import CatalogSchema
 from databricks.labs.ucx.hive_metastore.grants import (
@@ -38,31 +43,39 @@ from databricks.labs.ucx.hive_metastore.grants import (
     ComputeLocations,
     Grant,
     GrantsCrawler,
+    GrantOwnership,
     MigrateGrants,
     PrincipalACL,
 )
 from databricks.labs.ucx.hive_metastore.mapping import TableMapping
 from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationIndex
+from databricks.labs.ucx.hive_metastore.ownership import (
+    TableMigrationOwnership,
+    TableOwnership,
+    TableOwnershipGrantLoader,
+    DefaultSecurableOwnership,
+)
 from databricks.labs.ucx.hive_metastore.table_migrate import (
     TableMigrationStatusRefresher,
     TablesMigrator,
 )
 from databricks.labs.ucx.hive_metastore.table_move import TableMove
-from databricks.labs.ucx.hive_metastore.udfs import UdfsCrawler
+from databricks.labs.ucx.hive_metastore.udfs import UdfsCrawler, UdfOwnership
 from databricks.labs.ucx.hive_metastore.verification import VerifyHasCatalog, VerifyHasMetastore
 from databricks.labs.ucx.installer.workflows import DeployedWorkflows
 from databricks.labs.ucx.progress.install import VerifyProgressTracking
 from databricks.labs.ucx.source_code.graph import DependencyResolver
-from databricks.labs.ucx.source_code.jobs import WorkflowLinter
+from databricks.labs.ucx.source_code.linters.jobs import WorkflowLinter
 from databricks.labs.ucx.source_code.known import KnownList
-from databricks.labs.ucx.source_code.linters.files import FileLoader, FolderLoader, ImportFileResolver
+from databricks.labs.ucx.source_code.folders import FolderLoader
+from databricks.labs.ucx.source_code.files import FileLoader, ImportFileResolver
 from databricks.labs.ucx.source_code.notebooks.loaders import (
     NotebookResolver,
     NotebookLoader,
 )
 from databricks.labs.ucx.source_code.path_lookup import PathLookup
-from databricks.labs.ucx.source_code.queries import QueryLinter
-from databricks.labs.ucx.source_code.redash import Redash
+from databricks.labs.ucx.source_code.linters.queries import QueryLinter
+from databricks.labs.ucx.source_code.linters.redash import Redash
 from databricks.labs.ucx.workspace_access import generic, redash
 from databricks.labs.ucx.workspace_access.groups import GroupManager
 from databricks.labs.ucx.workspace_access.manager import PermissionManager
@@ -243,14 +256,98 @@ class GlobalContext(abc.ABC):
     def grants_crawler(self) -> GrantsCrawler:
         return GrantsCrawler(self.tables_crawler, self.udfs_crawler, self.config.include_databases)
 
+    @cached_property
+    def grant_ownership(self) -> GrantOwnership:
+        return GrantOwnership(self.administrator_locator)
+
     @cached_property
     def udfs_crawler(self) -> UdfsCrawler:
         return UdfsCrawler(self.sql_backend, self.inventory_database, self.config.include_databases)
 
+    @cached_property
+    def udf_ownership(self) -> UdfOwnership:
+        return UdfOwnership(self.administrator_locator)
+
+    @cached_property
+    def dashboard_ownership(self) -> DashboardOwnership:
+        return DashboardOwnership(self.administrator_locator, self.workspace_client, self.workspace_path_ownership)
+
     @cached_property
     def tables_crawler(self) -> TablesCrawler:
         return TablesCrawler(self.sql_backend, self.inventory_database, self.config.include_databases)
 
+    @cached_property
+    def jobs_crawler(self) -> JobsCrawler:
+        return JobsCrawler(
+            self.workspace_client,
+            self.sql_backend,
+            self.inventory_database,
+            include_job_ids=self.config.include_job_ids,
+            exclude_job_ids=[int(job_id) for job_id in self.install_state.jobs.values()],
+        )
+
+    @cached_property
+    def table_ownership(self) -> TableOwnership:
+        return TableOwnership(
+            self.administrator_locator,
+            self.grants_crawler,
+            self.used_tables_crawler_for_paths,
+            self.used_tables_crawler_for_queries,
+            self.legacy_query_ownership,
+            self.workspace_path_ownership,
+        )
+
+    @cached_property
+    def redash_crawler(self) -> RedashDashboardCrawler:
+        return RedashDashboardCrawler(
+            self.workspace_client,
+            self.sql_backend,
+            self.inventory_database,
+            include_dashboard_ids=self.config.include_dashboard_ids,
+            include_query_ids=self.config.include_query_ids,
+            debug_listing_upper_limit=self.config.debug_listing_upper_limit,
+        )
+
+    @cached_property
+    def lakeview_crawler(self) -> LakeviewDashboardCrawler:
+        return LakeviewDashboardCrawler(
+            self.workspace_client,
+            self.sql_backend,
+            self.inventory_database,
+            include_dashboard_ids=self.config.include_dashboard_ids,
+            exclude_dashboard_ids=list(self.install_state.dashboards.values()),
+            include_query_ids=self.config.include_query_ids,
+        )
+
+    @cached_property
+    def default_securable_ownership(self) -> DefaultSecurableOwnership:
+        # validate that the default_owner_group is set and is a valid group (the current user is a member)
+
+        return DefaultSecurableOwnership(
+            self.administrator_locator,
+            self.tables_crawler,
+            self.group_manager,
+            self.config.default_owner_group,
+            lambda: self.workspace_client.current_user.me().user_name,
+        )
+
+    @cached_property
+    def workspace_path_ownership(self) -> WorkspacePathOwnership:
+        return WorkspacePathOwnership(self.administrator_locator, self.workspace_client)
+
+    @cached_property
+    def legacy_query_ownership(self) -> LegacyQueryOwnership:
+        return LegacyQueryOwnership(self.administrator_locator, self.workspace_client)
+
+    @cached_property
+    def directfs_access_ownership(self) -> DirectFsAccessOwnership:
+        return DirectFsAccessOwnership(
+            self.administrator_locator,
+            self.workspace_path_ownership,
+            self.legacy_query_ownership,
+            self.workspace_client,
+        )
+
     @cached_property
     def tables_migrator(self) -> TablesMigrator:
         return TablesMigrator(
@@ -274,11 +371,40 @@ class GlobalContext(abc.ABC):
             self.workspace_info,
             self.migration_status_refresher,
             self.migrate_grants,
+            self.sql_backend,
+            self.config.inventory_database,
+        )
+
+    @cached_property
+    def table_ownership_grant_loader(self) -> TableOwnershipGrantLoader:
+        return TableOwnershipGrantLoader(self.tables_crawler, self.default_securable_ownership)
+
+    @cached_property
+    def pipelines_migrator(self) -> PipelinesMigrator:
+        include_pipeline_ids = (
+            self.named_parameters.get('include_pipeline_ids', '').split(',')
+            if 'include_pipeline_ids' in self.named_parameters
+            else None
+        )
+        exclude_pipeline_ids = (
+            self.named_parameters.get('exclude_pipeline_ids', '').split(',')
+            if 'exclude_pipeline_ids' in self.named_parameters
+            else None
+        )
+        return PipelinesMigrator(
+            self.workspace_client,
+            self.pipelines_crawler,
+            self.jobs_crawler,
+            self.config.ucx_catalog,
+            include_pipeline_ids=include_pipeline_ids,
+            exclude_pipeline_ids=exclude_pipeline_ids,
         )
 
     @cached_property
     def migrate_grants(self) -> MigrateGrants:
+        # owner grants have to come first
         grant_loaders: list[Callable[[], Iterable[Grant]]] = [
+            self.default_securable_ownership.load,
             self.grants_crawler.snapshot,
             self.principal_acl.get_interactive_cluster_grants,
         ]
@@ -286,6 +412,7 @@ class GlobalContext(abc.ABC):
             self.sql_backend,
             self.group_manager,
             grant_loaders,
+            skip_tacl_migration=self.config.skip_tacl_migration,
         )
 
     @cached_property
@@ -298,9 +425,12 @@ class GlobalContext(abc.ABC):
             self.sql_backend,
             self.workspace_client,
             self.inventory_database,
-            self.config.enable_hms_federation,
         )
 
+    @cached_property
+    def pipelines_crawler(self) -> PipelinesCrawler:
+        return PipelinesCrawler(self.workspace_client, self.sql_backend, self.inventory_database)
+
     @cached_property
     def azure_service_principal_crawler(self) -> AzureServicePrincipalCrawler:
         return AzureServicePrincipalCrawler(self.workspace_client, self.sql_backend, self.inventory_database)
@@ -313,6 +443,7 @@ class GlobalContext(abc.ABC):
             self.inventory_database,
             self.tables_crawler,
             self.mounts_crawler,
+            enable_hms_federation=self.config.enable_hms_federation,
         )
 
     @cached_property
@@ -363,6 +494,10 @@ class GlobalContext(abc.ABC):
             self.tables_crawler,
         )
 
+    @cached_property
+    def table_migration_ownership(self) -> TableMigrationOwnership:
+        return TableMigrationOwnership(self.tables_crawler, self.table_ownership)
+
     @cached_property
     def iam_credential_manager(self) -> CredentialManager:
         return CredentialManager(self.workspace_client)
@@ -409,7 +544,7 @@ class GlobalContext(abc.ABC):
 
     @cached_property
     def pip_resolver(self) -> PythonLibraryResolver:
-        return PythonLibraryResolver(self.allow_list)
+        return PythonLibraryResolver(allow_list=self.allow_list)
 
     @cached_property
     def notebook_loader(self) -> NotebookLoader:
@@ -443,7 +578,7 @@ class GlobalContext(abc.ABC):
 
     @cached_property
     def file_resolver(self) -> ImportFileResolver:
-        return ImportFileResolver(self.file_loader, self.allow_list)
+        return ImportFileResolver(self.file_loader, allow_list=self.allow_list)
 
     @cached_property
     def dependency_resolver(self) -> DependencyResolver:
@@ -455,22 +590,24 @@ class GlobalContext(abc.ABC):
     def workflow_linter(self) -> WorkflowLinter:
         return WorkflowLinter(
             self.workspace_client,
+            self.jobs_crawler,
             self.dependency_resolver,
             self.path_lookup,
             TableMigrationIndex([]),  # TODO: bring back self.tables_migrator.index()
             self.directfs_access_crawler_for_paths,
             self.used_tables_crawler_for_paths,
-            self.config.include_job_ids,
         )
 
     @cached_property
     def query_linter(self) -> QueryLinter:
         return QueryLinter(
-            self.workspace_client,
-            TableMigrationIndex([]),  # TODO: bring back self.tables_migrator.index()
+            self.sql_backend,
+            self.inventory_database,
+            TableMigrationIndex([]),
             self.directfs_access_crawler_for_queries,
             self.used_tables_crawler_for_queries,
-            self.config.include_dashboard_ids,
+            [self.redash_crawler, self.lakeview_crawler],
+            self.config.debug_listing_upper_limit,
         )
 
     @cached_property
@@ -495,6 +632,7 @@ class GlobalContext(abc.ABC):
             self.migration_status_refresher.index(),
             self.workspace_client,
             self.installation,
+            self.redash_crawler,
         )
 
     @cached_property
