@@ -3,7 +3,7 @@ from collections import defaultdict
 from collections.abc import Callable, Iterable
 from dataclasses import dataclass, replace
 from functools import partial, cached_property
-from typing import Protocol
+from typing import ClassVar, Protocol
 
 from databricks.labs.blueprint.installation import Installation
 from databricks.labs.blueprint.parallel import ManyError, Threads
@@ -66,6 +66,8 @@ class Grant:
     any_file: bool = False
     anonymous_function: bool = False
 
+    __id_attributes__: ClassVar[tuple[str, ...]] = ("object_type", "object_key", "action_type", "principal")
+
     @staticmethod
     def type_and_key(
         *,
@@ -105,6 +107,11 @@ class Grant:
         )
         raise ValueError(msg)
 
+    @property
+    def object_type(self) -> str:
+        this_type, _ = self.this_type_and_key()
+        return this_type
+
     @property
     def object_key(self) -> str:
         _, key = self.this_type_and_key()
@@ -112,12 +119,12 @@ class Grant:
 
     @property
     def order(self) -> int:
-        """Order of the grants to be upheld when applying."""
-        match self.action_type:
-            case "OWN":  # Apply ownership as last to avoid losing permissions for applying grants
-                return 1
-            case _:
-                return 0
+        """Apply Ownership First, then the rest.
+        Consider Revising When we properly apply manage permission"""
+
+        if self.action_type == "OWN":  # Apply ownership as last to avoid losing permissions for applying grants
+            return 0
+        return 1
 
     def this_type_and_key(self):
         return self.type_and_key(
@@ -182,14 +189,17 @@ class Grant:
             ("FUNCTION", "SELECT"): self._uc_action("EXECUTE"),
             ("TABLE", "SELECT"): self._uc_action("SELECT"),
             ("TABLE", "MODIFY"): self._uc_action("MODIFY"),
+            ("TABLE", "READ_METADATA"): self._uc_action("BROWSE"),
             ("TABLE", "ALL PRIVILEGES"): self._uc_action("ALL PRIVILEGES"),
             ("TABLE", "OWN"): self._set_owner_sql,
             ("VIEW", "SELECT"): self._uc_action("SELECT"),
+            ("VIEW", "READ_METADATA"): self._uc_action("BROWSE"),
             ("VIEW", "OWN"): self._set_owner_sql,
             ("DATABASE", "USAGE"): self._uc_action("USE SCHEMA"),
             ("DATABASE", "CREATE"): self._uc_action("CREATE TABLE"),
             ("DATABASE", "CREATE_NAMED_FUNCTION"): self._uc_action("CREATE FUNCTION"),
             ("DATABASE", "SELECT"): self._uc_action("SELECT"),
+            ("DATABASE", "READ_METADATA"): self._uc_action("BROWSE"),
             ("DATABASE", "MODIFY"): self._uc_action("MODIFY"),
             ("DATABASE", "OWN"): self._set_owner_sql,
             ("CATALOG", "OWN"): self._set_owner_sql,
@@ -208,12 +218,12 @@ CLUSTER_WITHOUT_ACL_FRAGMENT = "Table Access Control is not enabled on this clus
 class GrantsCrawler(CrawlerBase[Grant]):
     """Crawler that captures access controls that relate to data and other securable objects."""
 
-    def __init__(self, tc: TablesCrawler, udf: UdfsCrawler, include_databases: list[str] | None = None):
-        assert tc._backend == udf._backend
-        assert tc._catalog == udf._catalog
-        assert tc._schema == udf._schema
-        super().__init__(tc._backend, tc._catalog, tc._schema, "grants", Grant)
-        self._tc = tc
+    def __init__(self, tables_crawler: TablesCrawler, udf: UdfsCrawler, include_databases: list[str] | None = None):
+        assert tables_crawler._sql_backend == udf._sql_backend
+        assert tables_crawler._catalog == udf._catalog
+        assert tables_crawler._schema == udf._schema
+        super().__init__(tables_crawler._sql_backend, tables_crawler._catalog, tables_crawler._schema, "grants", Grant)
+        self._tables_crawler = tables_crawler
         self._udf = udf
         self._include_databases = include_databases
 
@@ -268,7 +278,7 @@ class GrantsCrawler(CrawlerBase[Grant]):
         else:
             for database in self._include_databases:
                 tasks.append(partial(self.grants, catalog=catalog, database=database))
-        for table in self._tc.snapshot():
+        for table in self._tables_crawler.snapshot():
             fn = partial(self.grants, catalog=catalog, database=table.database)
             # Views are treated as a type of table and enumerated by the table crawler.
             if table.view_text is None:
@@ -405,10 +415,10 @@ class AwsACL:
     def __init__(
         self,
         ws: WorkspaceClient,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         installation: Installation,
     ):
-        self._backend = backend
+        self._sql_backend = sql_backend
         self._ws = ws
         self._installation = installation
 
@@ -506,7 +516,7 @@ class AwsACL:
                 else:
                     resource_url = permission_mapping.resource_path
                 if location.url.startswith(resource_url) and permission_mapping.role_arn == role_name:
-                    matching_location[location.url] = permission_mapping.privilege
+                    matching_location[location.url] = permission_mapping.privilege.value
         return matching_location
 
 
@@ -514,12 +524,12 @@ class AzureACL:
     def __init__(
         self,
         ws: WorkspaceClient,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         spn_crawler: AzureServicePrincipalCrawler,
         installation: Installation,
     ):
-        self._backend = backend
         self._ws = ws
+        self._sql_backend = sql_backend
         self._spn_crawler = spn_crawler
         self._installation = installation
 
@@ -597,13 +607,13 @@ class PrincipalACL:
     def __init__(
         self,
         ws: WorkspaceClient,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         installation: Installation,
         tables_crawler: TablesCrawler,
         external_locations: ExternalLocations,
         cluster_locations: Callable[[], list[ComputeLocations]],
     ):
-        self._backend = backend
+        self._sql_backend = sql_backend
         self._ws = ws
         self._installation = installation
         self._tables_crawler = tables_crawler
@@ -727,11 +737,11 @@ class PrincipalACL:
                 self._update_location_permissions(location_name, principals)
         logger.info("Applied all the permission on external location")
 
-    def _update_location_permissions(self, location_name: str, principals: list[str]):
+    def _update_location_permissions(self, location_name: str, principals: list[str]) -> None:
         permissions = [Privilege.CREATE_EXTERNAL_TABLE, Privilege.CREATE_EXTERNAL_VOLUME, Privilege.READ_FILES]
         changes = [PermissionsChange(add=permissions, principal=principal) for principal in principals]
         self._ws.grants.update(
-            SecurableType.EXTERNAL_LOCATION,
+            SecurableType.EXTERNAL_LOCATION.value,
             location_name,
             changes=changes,
         )
@@ -770,13 +780,22 @@ class MigrateGrants:
         sql_backend: SqlBackend,
         group_manager: GroupManager,
         grant_loaders: list[Callable[[], Iterable[Grant]]],
+        *,
+        skip_tacl_migration: bool = False,
     ):
         self._sql_backend = sql_backend
         self._group_manager = group_manager
+        # List of grant loaders, the assumption that the first one is a loader for ownership grants
         self._grant_loaders = grant_loaders
+        if skip_tacl_migration:
+            logger.warning("Skipping TACL migration")
+        self._skip_tacl_migration = skip_tacl_migration
 
     def apply(self, src: SecurableObject, dst: SecurableObject) -> bool:
-        for grant in self._match_grants(src):
+        if self._skip_tacl_migration:
+            return True
+        grants = self._match_grants(src)
+        for grant in grants:
             acl_migrate_sql = grant.uc_grant_sql(dst.kind, dst.full_name)
             if acl_migrate_sql is None:
                 logger.warning(
@@ -793,15 +812,35 @@ class MigrateGrants:
                 )
         return True
 
+    def retrieve(self, src: SecurableObject, dst: SecurableObject) -> list[Grant]:
+        grants = self._match_grants(src)
+        discover_grants = []
+        for grant in grants:
+            acl_migrate_sql = grant.uc_grant_sql(dst.kind, dst.full_name)
+            if acl_migrate_sql is None:
+                logger.warning(
+                    f"failed-to-migrate: Hive metastore grant '{grant.action_type}' cannot be mapped to UC grant for "
+                    f"{dst.kind} '{dst.full_name}'. Skipping."
+                )
+                continue
+            logger.debug(f"Retrieving acls on {dst.full_name} using SQL query: {acl_migrate_sql}")
+            discover_grants.append(grant)
+        return discover_grants
+
     @cached_property
     def _workspace_to_account_group_names(self) -> dict[str, str]:
         return {g.name_in_workspace: g.name_in_account for g in self._group_manager.snapshot()}
 
     @cached_property
     def _grants(self) -> list[Grant]:
+        # Accumulate grants from all loaders skips ownership grants
         grants = []
-        for loader in self._grant_loaders:
+        for index, loader in enumerate(self._grant_loaders):
             for grant in loader():
+                # Skip ownership grants for all loaders other than the first one
+                # The assumption is that the first one will be designated as the ownership loader
+                if index != 0 and grant.action_type == "OWN":
+                    continue
                 grants.append(grant)
         return grants
 
@@ -820,23 +859,46 @@ class MigrateGrants:
             return grant
         return replace(grant, principal=target_principal)
 
+    def _apply_ownership(self, dst: SecurableObject, owner: str):
+        owner = escape_sql_identifier(owner)
+        destination = escape_sql_identifier(dst.full_name)
+        if dst.kind == "TABLE":
+            sql = f"ALTER TABLE {destination} " f"SET OWNER TO {owner}"
+        elif dst.kind == "VIEW":
+            sql = f"ALTER VIEW {destination} " f"SET OWNER TO {owner}"
+        elif dst.kind in {"SCHEMA", "DATABASE"}:
+            sql = f"ALTER SCHEMA {destination} " f"SET OWNER TO {owner}"
+        elif dst.kind == "CATALOG":
+            sql = f"ALTER CATALOG {destination} " f"SET OWNER TO {owner}"
+        else:
+            logger.warning(f"Unknown object type {dst.kind} for ownership migration")
+            return
+        logger.debug(f"Applying ownership on {dst.full_name} using SQL query: {sql}")
+        try:
+            self._sql_backend.execute(sql)
+        except DatabricksError as e:
+            logger.warning(f"Failed to apply ownership on {dst.full_name}", exc_info=e)
 
-class ACLMigrator:
+
+class ACLMigrator(CrawlerBase[Grant]):
     def __init__(
         self,
         tables_crawler: TablesCrawler,
         workspace_info: WorkspaceInfo,
         migration_status_refresher: TableMigrationStatusRefresher,
         migrate_grants: MigrateGrants,
+        backend: SqlBackend,
+        schema: str,
     ):
-        self._table_crawler = tables_crawler
+        self._tables_crawler = tables_crawler
         self._workspace_info = workspace_info
         self._migration_status_refresher = migration_status_refresher
         self._migrate_grants = migrate_grants
+        super().__init__(backend, "hive_metastore", schema, "inferred_grants", Grant)
 
     def migrate_acls(self, *, target_catalog: str | None = None, hms_fed: bool = False) -> None:
         workspace_name = self._workspace_info.current()
-        tables = list(self._table_crawler.snapshot())
+        tables = list(self._tables_crawler.snapshot())
         if not tables:
             logger.info("No tables found to acl")
             return
@@ -846,6 +908,20 @@ class ACLMigrator:
             tables_to_migrate = self._get_migrated_tables(tables)
         self._migrate_acls(tables_to_migrate)
 
+    def _retrieve_table_acls(self, *, target_catalog: str | None = None, hms_fed: bool = False) -> Iterable[Grant]:
+        tables = list(self._tables_crawler.snapshot())
+        grants: list[Grant] = []
+        if not tables:
+            logger.info("No tables found to acl")
+            return grants
+        if hms_fed:
+            tables_to_migrate = self._get_hms_fed_tables(
+                tables, target_catalog if target_catalog else self._workspace_info.current()
+            )
+        else:
+            tables_to_migrate = self._get_migrated_tables(tables)
+        return self._retrieve_acls(tables_to_migrate)
+
     def _get_migrated_tables(self, tables: list[Table]) -> list[TableToMigrate]:
         # gets all the migrated table to apply ACLs to
         tables_to_migrate = []
@@ -892,9 +968,18 @@ class ACLMigrator:
     def _migrate_acls(self, tables_in_scope: list[TableToMigrate]) -> None:
         tasks = []
         for table in tables_in_scope:
-            tasks.append(partial(self._migrate_grants.apply, table.src, table.rule.as_uc_table_key))
+            tasks.append(partial(self._migrate_grants.apply, table.src, table.rule.as_uc_table))
         Threads.strict("migrate grants", tasks)
 
-    def _is_migrated(self, schema: str, table: str) -> bool:
-        index = self._migration_status_refresher.index()
-        return index.is_migrated(schema, table)
+    def _retrieve_acls(self, tables_in_scope: list[TableToMigrate]) -> Iterable[Grant]:
+        grants = []
+        for table in tables_in_scope:
+            grants += self._migrate_grants.retrieve(table.src, table.rule.as_uc_table)
+        return grants
+
+    def _crawl(self) -> Iterable[Grant]:
+        return self._retrieve_table_acls()
+
+    def _try_fetch(self):
+        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
+            yield Grant(*row)
