@@ -20,7 +20,6 @@ from databricks.labs.ucx.hive_metastore.tables import Table, TablesCrawler
 
 logger = logging.getLogger(__name__)
 
-
 _EXTERNAL_FILE_LOCATION_SCHEMES = ("s3", "s3a", "s3n", "gcs", "abfss")
 
 
@@ -70,7 +69,8 @@ class LocationTrie:
     def _parse_location(cls, location: str | None) -> list[str]:
         if not location:
             return []
-        parse_result = cls._parse_url(location.rstrip("/"))
+        location = ExternalLocations.clean_location(location)
+        parse_result = cls._parse_url(location)
         if not parse_result:
             return []
         parts = [parse_result.scheme, parse_result.netloc]
@@ -151,21 +151,70 @@ class ExternalLocations(CrawlerBase[ExternalLocation]):
     def __init__(
         self,
         ws: WorkspaceClient,
-        sbe: SqlBackend,
+        sql_backend: SqlBackend,
         schema: str,
         tables_crawler: TablesCrawler,
         mounts_crawler: 'MountsCrawler',
+        *,
+        enable_hms_federation: bool = False,
     ):
-        super().__init__(sbe, "hive_metastore", schema, "external_locations", ExternalLocation)
+        super().__init__(sql_backend, "hive_metastore", schema, "external_locations", ExternalLocation)
         self._ws = ws
         self._tables_crawler = tables_crawler
         self._mounts_crawler = mounts_crawler
+        self._enable_hms_federation = enable_hms_federation
 
     @cached_property
     def _mounts_snapshot(self) -> list['Mount']:
         """Returns all mounts, sorted by longest prefixes first."""
         return sorted(self._mounts_crawler.snapshot(), key=lambda _: (len(_.name), _.name), reverse=True)
 
+    @staticmethod
+    def clean_location(location: str) -> str:
+        # remove the s3a scheme and replace it with s3 as these can be considered the same and will be treated as such
+        # Having s3a and s3 as separate locations will cause issues when trying to find overlapping locations
+        return re.sub(r"^s3a:/", r"s3:/", location).rstrip("/")
+
+    def external_locations_with_root(self) -> Iterable[ExternalLocation]:
+        """
+        Produces a list of external locations with the DBFS root location appended to the list.
+        Utilizes the snapshot method.
+        Used for HMS Federation.
+
+        Yields:
+                Iterable[Result]: Combination of all the external locations and the DBFS root location
+        """
+
+        yield from self.snapshot()
+        dbfs_root = self._get_dbfs_root()
+        if dbfs_root:
+            yield dbfs_root
+
+    def _get_dbfs_root(self) -> ExternalLocation | None:
+        """
+        Get the root location of the DBFS only if HMS Fed is enabled.
+        Utilizes an undocumented Databricks API call
+
+        Returns:
+            Cloud storage root location for dbfs
+
+        """
+        if not self._enable_hms_federation:
+            return None
+        logger.debug("Retrieving DBFS root location")
+        try:
+            response = self._ws.api_client.do("GET", "/api/2.0/dbfs/resolve-path", query={"path": "dbfs:/"})
+            if isinstance(response, dict):
+                resolved_path = response.get("resolved_path")
+                if resolved_path:
+                    path = f"{self.clean_location(resolved_path)}/user/hive/warehouse"
+                    return ExternalLocation(path, 0)
+        except NotFound:
+            # Couldn't retrieve the DBFS root location
+            logger.warning("DBFS root location not found")
+            return None
+        return None
+
     def _external_locations(self) -> Iterable[ExternalLocation]:
         trie = LocationTrie()
         for table in self._tables_crawler.snapshot():
@@ -179,9 +228,18 @@ class ExternalLocations(CrawlerBase[ExternalLocation]):
             curr = queue.pop()
             num_children = len(curr.children)  # 0 - take parent
             if curr.location and (num_children > 1 or num_children == 0):
-                if curr.parent and num_children == 0 and not curr.is_jdbc():  # one table having the prefix
+                # Checking if the parent location is a valid location for external location.
+                # If the table location is a leaf location (a foldermore than 2 levels from the root)
+                # the parent folder will be considered as external location.
+                # If the table location is a root location (a folder at the root level) it will be considered as
+                # external location.
+
+                if (
+                    curr.parent and curr.parent.is_valid() and num_children == 0 and not curr.is_jdbc()
+                ):  # one table having the prefix
                     curr = curr.parent
-                assert curr.location is not None
+                if not curr.location:
+                    continue
                 external_location = ExternalLocation(curr.location, len(list(curr.all_tables())))
                 external_locations.append(external_location)
                 continue
@@ -345,14 +403,12 @@ class Mount:
 class MountsCrawler(CrawlerBase[Mount]):
     def __init__(
         self,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         ws: WorkspaceClient,
         inventory_database: str,
-        enable_hms_federation: bool = False,
     ):
-        super().__init__(backend, "hive_metastore", inventory_database, "mounts", Mount)
+        super().__init__(sql_backend, "hive_metastore", inventory_database, "mounts", Mount)
         self._dbutils = ws.dbutils
-        self._enable_hms_federation = enable_hms_federation
 
     @staticmethod
     def _deduplicate_mounts(mounts: list) -> list:
@@ -368,48 +424,11 @@ class MountsCrawler(CrawlerBase[Mount]):
                 deduplicated_mounts.append(obj)
         return deduplicated_mounts
 
-    @cached_property
-    def _jvm(self):
-        # pylint: disable=import-error,import-outside-toplevel,broad-exception-caught
-        try:
-            from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]
-
-            spark = SparkSession.builder.getOrCreate()
-            return spark._jvm  # pylint: disable=protected-access
-        except Exception as err:
-            logger.warning(f"Cannot create Py4j proxy: {err}")
-            return None
-
-    def _resolve_dbfs_root(self) -> Mount | None:
-        # pylint: disable=broad-exception-caught,too-many-try-statements
-        try:
-            jvm = self._jvm
-            if not jvm:
-                return None
-            uri = jvm.java.net.URI
-            some = jvm.scala.Some
-            hms_fed_dbfs_utils = jvm.com.databricks.sql.managedcatalog.connections.HmsFedDbfsUtils
-            root_location_opt = hms_fed_dbfs_utils.resolveDbfsPath(some(uri("dbfs:/user/hive/warehouse")))
-            if root_location_opt.isDefined():
-                source: str = root_location_opt.get().toString()
-                source = source.removesuffix('user/hive/warehouse')
-                return Mount("/", source)
-            return None
-        except Exception as err:
-            logger.warning(f"Failed to resolve DBFS root location: {err}")
-            return None
-
     def _crawl(self) -> Iterable[Mount]:
         mounts = []
         try:
             for mount_point, source, _ in self._dbutils.fs.mounts():
                 mounts.append(Mount(mount_point, source))
-            if self._enable_hms_federation:
-                root_mount = self._resolve_dbfs_root()
-                if root_mount:
-                    # filter out DatabricksRoot, otherwise ExternalLocations.resolve_mount() won't work
-                    mounts = list(filter(lambda _: _.source != 'DatabricksRoot', mounts))
-                    mounts.append(root_mount)
         except Exception as error:  # pylint: disable=broad-except
             if "com.databricks.backend.daemon.dbutils.DBUtilsCore.mounts() is not whitelisted" in str(error):
                 logger.warning(
@@ -446,7 +465,7 @@ class TablesInMounts(CrawlerBase[Table]):
 
     def __init__(
         self,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         ws: WorkspaceClient,
         inventory_database: str,
         mounts_crawler: MountsCrawler,
@@ -454,7 +473,7 @@ class TablesInMounts(CrawlerBase[Table]):
         exclude_paths_in_mount: list[str] | None = None,
         include_paths_in_mount: list[str] | None = None,
     ):
-        super().__init__(backend, "hive_metastore", inventory_database, "tables", Table)
+        super().__init__(sql_backend, "hive_metastore", inventory_database, "tables", Table)
         self._dbutils = ws.dbutils
         self._mounts_crawler = mounts_crawler
         self._include_mounts = include_mounts
@@ -566,46 +585,71 @@ class TablesInMounts(CrawlerBase[Table]):
         if delta_log_folders is None:
             delta_log_folders = {}
         logger.info(f"Listing {root_dir}")
-        file_infos = self._dbutils.fs.ls(root_dir)
+        file_infos = self._dbutils.fs.ls(root_dir) or []
         for file_info in file_infos:
             if self._is_irrelevant(file_info.name) or file_info.path == root_dir:
                 logger.debug(f"Path {file_info.path} is irrelevant")
                 continue
 
             root_path = os.path.dirname(root_dir)
-            previous_entry = delta_log_folders.get(root_path)
+            parent_entry = delta_log_folders.get(root_path)
             table_in_mount = self._assess_path(file_info)
 
-            if previous_entry:
-                # Happens when first folder was _delta_log and next folders are partitioned folder
-                if previous_entry.format == "DELTA" and self._is_partitioned(file_info.name):
-                    delta_log_folders[root_path] = TableInMount(format=previous_entry.format, is_partitioned=True)
-                # Happens when previous entries where partitioned folders and the current one is delta_log
-                if previous_entry.is_partitioned and table_in_mount and table_in_mount.format == "DELTA":
-                    delta_log_folders[root_path] = TableInMount(format=table_in_mount.format, is_partitioned=True)
-                continue
+            if parent_entry:
+                # if the root_path was already found to be partitioned, we can break to reduce redundant ops
+                if delta_log_folders[root_path] == TableInMount(format="DELTA", is_partitioned=True):
+                    break
 
-            if self._is_partitioned(file_info.name):
-                partition_format = self._find_partition_file_format(file_info.path)
-                if partition_format:
-                    delta_log_folders[root_path] = partition_format
-                continue
+                if (
+                    # Happens when first folder was _delta_log and next folders are partitioned folder
+                    parent_entry.format == "DELTA"
+                    and self._is_partitioned(file_info.name)
+                ) or (
+                    # Happens when previous entries where partitioned folders and the current one is delta_log
+                    parent_entry.is_partitioned
+                    and table_in_mount
+                    and table_in_mount.format == "DELTA"
+                ):
+                    delta_log_folders[root_path] = TableInMount(format="DELTA", is_partitioned=True)
+                    logger.debug(f"Added DELTA table for {root_path} (partitioned delta)")
 
-            if not table_in_mount:
-                self._find_delta_log_folders(file_info.path, delta_log_folders)
-                continue
+                self._recurse_dir_if_needed(file_info, delta_log_folders)
+
+            elif self._is_partitioned(file_info.name):
+                self._handle_partitioned_file(file_info.path, root_path, delta_log_folders)
+
+            elif not table_in_mount:
+                self._recurse_dir_if_needed(file_info, delta_log_folders)
+
+            elif table_in_mount.format == "DELTA" and file_info.name == "_delta_log/":
+                delta_log_folders[root_path] = table_in_mount
+                logger.debug(f"Added {table_in_mount.format} table for {root_path} (normal delta)")
+
+            else:
+                delta_log_folders[root_path] = table_in_mount
+                logger.debug(f"Added {table_in_mount.format} table for {root_path} (general)")
+                self._recurse_dir_if_needed(file_info, delta_log_folders)
 
-            delta_log_folders[root_path] = table_in_mount
         return delta_log_folders
 
+    def _recurse_dir_if_needed(self, file_info: FileInfo, delta_log_folders: dict[str, TableInMount]) -> None:
+        if self._is_recursible_dir(file_info):
+            self._find_delta_log_folders(file_info.path, delta_log_folders)
+
+    def _handle_partitioned_file(self, file_path, root_path, delta_log_folders):
+        partition_format = self._find_partition_file_format(file_path)
+        if partition_format:
+            delta_log_folders[root_path] = partition_format
+            logger.debug(f"Added {partition_format.format} table for {root_path} (partitioned)")
+
     def _find_partition_file_format(self, root_dir: str) -> TableInMount | None:
-        logger.info(f"Listing {root_dir}")
+        logger.info(f"Listing partitioned file {root_dir}")
         file_infos = self._dbutils.fs.ls(root_dir)
         for file_info in file_infos:
             path_extension = self._assess_path(file_info)
             if path_extension:
                 return TableInMount(format=path_extension.format, is_partitioned=True)
-            if self._is_partitioned(file_info.name):
+            if self._is_partitioned(file_info.name) and file_info.path != root_dir:
                 return self._find_partition_file_format(file_info.path)
         return None
 
@@ -624,6 +668,23 @@ class TablesInMounts(CrawlerBase[Table]):
             return TableInMount(format="PARQUET", is_partitioned=False)
         return None
 
+    def _is_recursible_dir(self, file_info: FileInfo) -> bool:
+        # Rules for recursing into a folder
+        # - should be size 0 (usually a directory)
+        # - should not be the _delta_log directory
+        # - file should not be partitioned
+        # - file name should not start with 'part-'
+        # - no brackets, brackets are not allowed in dbutils.fs.ls
+        # - is not a streaming checkpoint dir
+        return (
+            file_info.size == 0
+            and file_info.name != "_delta_log/"
+            and not self._is_partitioned(file_info.name)
+            and not file_info.name.startswith("part-")
+            and not any(char in file_info.name for char in "[]")
+            and not self._is_streaming_checkpoint_dir(file_info)
+        )
+
     @staticmethod
     def _is_partitioned(file_name: str) -> bool:
         return '=' in file_name
@@ -645,3 +706,7 @@ class TablesInMounts(CrawlerBase[Table]):
 
     def _is_irrelevant(self, file_name: str) -> bool:
         return any(pattern in file_name for pattern in self._fiter_paths)
+
+    def _is_streaming_checkpoint_dir(self, file_info: FileInfo) -> bool:
+        path_dirs = [file.name for file in self._dbutils.fs.ls(file_info.path) if file.size == 0]
+        return 'commits/' in path_dirs and 'offsets/' in path_dirs
