@@ -2,6 +2,7 @@ import dataclasses
 import logging
 import re
 from collections import defaultdict
+from collections.abc import Iterable
 from functools import partial, cached_property
 
 from databricks.labs.blueprint.parallel import Threads
@@ -18,8 +19,11 @@ from databricks.labs.ucx.hive_metastore.mapping import (
     TableMapping,
     TableToMigrate,
 )
-
-from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatusRefresher
+from databricks.labs.ucx.hive_metastore.table_migration_status import (
+    TableMigrationStatusRefresher,
+    TableMigrationStatus,
+    TableMigrationIndex,
+)
 from databricks.labs.ucx.hive_metastore.tables import (
     MigrationCount,
     Table,
@@ -31,47 +35,73 @@ from databricks.labs.ucx.hive_metastore.view_migrate import (
     ViewToMigrate,
 )
 
+
 logger = logging.getLogger(__name__)
 
 
 class TablesMigrator:
     def __init__(
         self,
-        table_crawler: TablesCrawler,
+        tables_crawler: TablesCrawler,
         ws: WorkspaceClient,
-        backend: SqlBackend,
+        sql_backend: SqlBackend,
         table_mapping: TableMapping,
         migration_status_refresher: TableMigrationStatusRefresher,
         migrate_grants: MigrateGrants,
         external_locations: ExternalLocations,
     ):
 
-        self._tc = table_crawler
-        self._backend = backend
+        self._tables_crawler = tables_crawler
+        self._sql_backend = sql_backend
         self._ws = ws
-        self._tm = table_mapping
+        self._table_mapping = table_mapping
         self._migration_status_refresher = migration_status_refresher
         self._seen_tables: dict[str, str] = {}
         self._migrate_grants = migrate_grants
         self._external_locations = external_locations
 
-    def get_remaining_tables(self) -> list[Table]:
-        self.index(force_refresh=True)
-        table_rows = []
-        for crawled_table in self._tc.snapshot():
-            if not self._is_migrated(crawled_table.database, crawled_table.name):
-                table_rows.append(crawled_table)
+    def warn_about_remaining_non_migrated_tables(self, migration_statuses: Iterable[TableMigrationStatus]) -> None:
+        migration_index = TableMigrationIndex(migration_statuses)
+        for crawled_table in self._tables_crawler.snapshot():
+            if not migration_index.is_migrated(crawled_table.database, crawled_table.name):
                 logger.warning(f"remained-hive-metastore-table: {crawled_table.key}")
-        return table_rows
 
     def index(self, *, force_refresh: bool = False):
         return self._migration_status_refresher.index(force_refresh=force_refresh)
 
+    def convert_managed_hms_to_external(
+        self,
+        managed_table_external_storage: str = "CLONE",
+    ):
+        # This method contains some of the steps of migrate tables. this was done to separate out the
+        # code for converting managed hms table to external, since this needs to run in non uc cluster,
+        # the functionality to call the UC api are removed
+
+        if managed_table_external_storage != "CONVERT_TO_EXTERNAL":
+            logger.info("Not required to convert managed hms table to external, Skipping this task...")
+            return None
+        self._spark = self._spark_session
+        tables_to_migrate = self._table_mapping.get_tables_to_migrate(self._tables_crawler, False)
+        tables_in_scope = filter(lambda t: t.src.what == What.EXTERNAL_SYNC, tables_to_migrate)
+        tasks = []
+        for table in tables_in_scope:
+            tasks.append(
+                partial(
+                    self._convert_hms_table_to_external,
+                    table.src,
+                )
+            )
+        Threads.strict("convert tables", tasks)
+        if not tasks:
+            logger.info("No managed hms table found to convert to external")
+        return tasks
+
     def migrate_tables(
         self,
         what: What,
         hiveserde_in_place_migrate: bool = False,
         managed_table_external_storage: str = "CLONE",
+        check_uc_table: bool = True,
     ):
         if managed_table_external_storage == "CONVERT_TO_EXTERNAL":
             self._spark = self._spark_session
@@ -82,9 +112,7 @@ class TablesMigrator:
         if what == What.VIEW:
             return self._migrate_views()
         return self._migrate_tables(
-            what,
-            managed_table_external_storage.upper(),
-            hiveserde_in_place_migrate,
+            what, managed_table_external_storage.upper(), hiveserde_in_place_migrate, check_uc_table
         )
 
     def _migrate_tables(
@@ -92,8 +120,9 @@ class TablesMigrator:
         what: What,
         managed_table_external_storage: str,
         hiveserde_in_place_migrate: bool = False,
+        check_uc_table: bool = True,
     ):
-        tables_to_migrate = self._tm.get_tables_to_migrate(self._tc)
+        tables_to_migrate = self._table_mapping.get_tables_to_migrate(self._tables_crawler, check_uc_table)
         tables_in_scope = filter(lambda t: t.src.what == what, tables_to_migrate)
         tasks = []
         for table in tables_in_scope:
@@ -112,19 +141,20 @@ class TablesMigrator:
         return tasks
 
     def _migrate_views(self):
-        tables_to_migrate = self._tm.get_tables_to_migrate(self._tc)
+        tables_to_migrate = self._table_mapping.get_tables_to_migrate(self._tables_crawler)
         all_tasks = []
         # Every batch of views to migrate needs an up-to-date table migration index
         # to determine if the dependencies have been migrated
-        sequencer = ViewsMigrationSequencer(tables_to_migrate, migration_index=self.index(force_refresh=True))
+        migration_index = self.index(force_refresh=True)
+        sequencer = ViewsMigrationSequencer(tables_to_migrate, migration_index=migration_index)
         batches = sequencer.sequence_batches()
         for batch in batches:
-            tasks = []
-            for view in batch:
-                tasks.append(partial(self._migrate_view, view))
+            # Avoid redundant refresh for first batch.
+            if all_tasks:
+                migration_index = self.index(force_refresh=True)
+            tasks = [partial(self._migrate_view, view, migration_index) for view in batch]
             Threads.strict("migrate views", tasks)
             all_tasks.extend(tasks)
-            self.index(force_refresh=True)
         return all_tasks
 
     @cached_property
@@ -134,12 +164,15 @@ class TablesMigrator:
 
         return SparkSession.builder.getOrCreate()
 
-    def _migrate_managed_table(self, managed_table_external_storage: str, src_table: TableToMigrate):
+    def _migrate_managed_table(
+        self,
+        managed_table_external_storage: str,
+        src_table: TableToMigrate,
+    ):
         if managed_table_external_storage == 'CONVERT_TO_EXTERNAL':
-            if self._convert_hms_table_to_external(src_table.src):
-                return self._migrate_external_table(
-                    src_table.src, src_table.rule
-                )  # _migrate_external_table remains unchanged
+            return self._migrate_external_table(
+                src_table.src, src_table.rule
+            )  # _migrate_external_table remains unchanged
         if managed_table_external_storage == 'SYNC_AS_EXTERNAL':
             return self._migrate_managed_as_external_table(src_table.src, src_table.rule)  # new method
         if managed_table_external_storage == 'CLONE':
@@ -204,30 +237,30 @@ class TablesMigrator:
         logger.info(f"Table {src_table.src.key} is not supported for migration")
         return True
 
-    def _migrate_view(self, src_view: ViewToMigrate):
+    def _migrate_view(self, src_view: ViewToMigrate, migration_index: TableMigrationIndex) -> bool:
         if self._table_already_migrated(src_view.rule.as_uc_table_key):
             logger.info(f"View {src_view.src.key} already migrated to {src_view.rule.as_uc_table_key}")
             return True
-        if self._view_can_be_migrated(src_view):
-            return self._migrate_view_table(src_view)
+        if self._view_can_be_migrated(src_view, migration_index):
+            return self._migrate_view_table(src_view, migration_index)
         logger.info(f"View {src_view.src.key} is not supported for migration")
         return True
 
-    def _view_can_be_migrated(self, view: ViewToMigrate):
+    def _view_can_be_migrated(self, view: ViewToMigrate, migration_index: TableMigrationIndex) -> bool:
         # dependencies have already been computed, therefore an empty dict is good enough
         for table in view.dependencies:
-            if not self.index().get(table.schema, table.name):
+            if not migration_index.get(table.schema, table.name):
                 logger.info(f"View {view.src.key} cannot be migrated because {table.key} is not migrated yet")
                 return False
         return True
 
-    def _migrate_view_table(self, src_view: ViewToMigrate):
-        view_migrate_sql = self._sql_migrate_view(src_view)
+    def _migrate_view_table(self, src_view: ViewToMigrate, migration_index: TableMigrationIndex):
+        view_migrate_sql = self._sql_migrate_view(src_view, migration_index)
         logger.debug(f"Migrating view {src_view.src.key} to using SQL query: {view_migrate_sql}")
         try:
-            self._backend.execute(view_migrate_sql)
-            self._backend.execute(self._sql_alter_to(src_view.src, src_view.rule.as_uc_table_key))
-            self._backend.execute(
+            self._sql_backend.execute(view_migrate_sql)
+            self._sql_backend.execute(self._sql_alter_to(src_view.src, src_view.rule.as_uc_table_key))
+            self._sql_backend.execute(
                 self._sql_alter_from(src_view.src, src_view.rule.as_uc_table_key, self._ws.get_workspace_id())
             )
         except DatabricksError as e:
@@ -235,13 +268,13 @@ class TablesMigrator:
             return False
         return self._migrate_grants.apply(src_view.src, src_view.rule.as_uc_table)
 
-    def _sql_migrate_view(self, src_view: ViewToMigrate) -> str:
+    def _sql_migrate_view(self, src_view: ViewToMigrate, migration_index: TableMigrationIndex) -> str:
         # We have to fetch create statement this way because of columns in:
         # CREATE VIEW x.y (col1, col2) AS SELECT * FROM w.t
-        create_statement = self._backend.fetch(f"SHOW CREATE TABLE {src_view.src.safe_sql_key}")
+        create_statement = self._sql_backend.fetch(f"SHOW CREATE TABLE {src_view.src.safe_sql_key}")
         src_view.src.view_text = next(iter(create_statement))["createtab_stmt"]
         # this does not require the index to be refreshed because the dependencies have already been validated
-        return src_view.sql_migrate_view(self.index())
+        return src_view.sql_migrate_view(migration_index)
 
     @cached_property
     def _catalog(self):
@@ -261,12 +294,28 @@ class TablesMigrator:
     def _catalog_table(self):
         return self._spark._jvm.org.apache.spark.sql.catalyst.catalog.CatalogTable  # pylint: disable=protected-access
 
-    def _convert_hms_table_to_external(self, src_table: Table):
+    @staticmethod
+    def _get_entity_storage_locations(table_metadata):
+        """Obtain the entityStorageLocations property for table metadata, if the property is present."""
+        # This is needed because:
+        #  - DBR 16.0 introduced entityStorageLocations as a property on table metadata, and this is required for
+        #    as a constructor parameter for CatalogTable.
+        #  - We need to be compatible with earlier versions of DBR.
+        #  - The normal hasattr() check does not work with Py4J-based objects: it always returns True and non-existent
+        #    methods will be automatically created on the proxy but fail when invoked.
+        # Instead the only approach is to use dir() to check if the method exists _prior_ to trying to access it.
+        # (After trying to access it, dir() will also include it even though it doesn't exist.)
+        return table_metadata.entityStorageLocations() if 'entityStorageLocations' in dir(table_metadata) else None
+
+    def _convert_hms_table_to_external(self, src_table: Table) -> bool:
+        """Converts a Hive metastore table to external using Spark JVM methods."""
+        logger.info(f"Changing HMS managed table {src_table.name} to External Table type.")
+        inventory_table = self._tables_crawler.full_name
         try:
-            logger.info(f"Changing HMS managed table {src_table.name} to External Table type.")
             database = self._spark._jvm.scala.Some(src_table.database)  # pylint: disable=protected-access
             table_identifier = self._table_identifier(src_table.name, database)
             old_table = self._catalog.getTableMetadata(table_identifier)
+            entity_storage_locations = self._get_entity_storage_locations(old_table)
             new_table = self._catalog_table(
                 old_table.identifier(),
                 self._catalog_type('EXTERNAL'),
@@ -288,27 +337,36 @@ class TablesMigrator:
                 old_table.schemaPreservesCase(),
                 old_table.ignoredProperties(),
                 old_table.viewOriginalText(),
+                # From DBR 16, there's a new constructor argument: entityStorageLocations (Seq[EntityStorageLocation])
+                # (We can't detect whether the argument is needed by the constructor, but assume that if the accessor
+                # is present on the source table then the argument is needed.)
+                *([entity_storage_locations] if entity_storage_locations is not None else []),
             )
             self._catalog.alterTable(new_table)
-            logger.info(f"Converted {src_table.name} to External Table type.")
+            self._update_table_status(src_table, inventory_table)
         except Exception as e:  # pylint: disable=broad-exception-caught
             logger.warning(f"Error converting HMS table {src_table.name} to external: {e}", exc_info=True)
             return False
+        logger.info(f"Converted {src_table.name} to External Table type.")
         return True
 
+    def _update_table_status(self, src_table: Table, inventory_table: str):
+        update_sql = f"UPDATE {escape_sql_identifier(inventory_table)} SET object_type = 'EXTERNAL' WHERE catalog='hive_metastore' AND database='{src_table.database}' AND name='{src_table.name}';"
+        self._sql_backend.execute(update_sql)
+
     def _migrate_managed_as_external_table(self, src_table: Table, rule: Rule):
         target_table_key = rule.as_uc_table_key
         table_migrate_sql = src_table.sql_migrate_as_external(target_table_key)
         logger.debug(f"Migrating external table {src_table.key} to using SQL query: {table_migrate_sql}")
         # have to wrap the fetch result with iter() for now, because StatementExecutionBackend returns iterator but RuntimeBackend returns list.
-        sync_result = next(iter(self._backend.fetch(table_migrate_sql)))
+        sync_result = next(iter(self._sql_backend.fetch(table_migrate_sql)))
         if sync_result.status_code != "SUCCESS":
             logger.warning(
                 f"failed-to-migrate: SYNC command failed to migrate table {src_table.key} to {target_table_key}. "
                 f"Status code: {sync_result.status_code}. Description: {sync_result.description}"
             )
             return False
-        self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+        self._sql_backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
         return self._migrate_grants.apply(src_table, rule.as_uc_table)
 
     def _migrate_external_table(self, src_table: Table, rule: Rule):
@@ -316,19 +374,19 @@ class TablesMigrator:
         table_migrate_sql = src_table.sql_migrate_external(target_table_key)
         logger.debug(f"Migrating external table {src_table.key} to using SQL query: {table_migrate_sql}")
         # have to wrap the fetch result with iter() for now, because StatementExecutionBackend returns iterator but RuntimeBackend returns list.
-        sync_result = next(iter(self._backend.fetch(table_migrate_sql)))
+        sync_result = next(iter(self._sql_backend.fetch(table_migrate_sql)))
         if sync_result.status_code != "SUCCESS":
             logger.warning(
                 f"failed-to-migrate: SYNC command failed to migrate table {src_table.key} to {target_table_key}. "
                 f"Status code: {sync_result.status_code}. Description: {sync_result.description}"
             )
             return False
-        self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+        self._sql_backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
         return self._migrate_grants.apply(src_table, rule.as_uc_table)
 
     def _migrate_external_table_hiveserde_in_place(self, src_table: Table, rule: Rule):
         # verify hive serde type
-        hiveserde_type = src_table.hiveserde_type(self._backend)
+        hiveserde_type = src_table.hiveserde_type(self._sql_backend)
         if hiveserde_type in [
             HiveSerdeType.NOT_HIVESERDE,
             HiveSerdeType.OTHER_HIVESERDE,
@@ -343,7 +401,7 @@ class TablesMigrator:
             dst_table_location = self._external_locations.resolve_mount(src_table.location)
 
         table_migrate_sql = src_table.sql_migrate_external_hiveserde_in_place(
-            rule.catalog_name, rule.dst_schema, rule.dst_table, self._backend, hiveserde_type, dst_table_location
+            rule.catalog_name, rule.dst_schema, rule.dst_table, self._sql_backend, hiveserde_type, dst_table_location
         )
         if not table_migrate_sql:
             logger.error(
@@ -355,10 +413,12 @@ class TablesMigrator:
             f"Migrating external table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
         )
         try:
-            self._backend.execute(table_migrate_sql)
-            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+            self._sql_backend.execute(table_migrate_sql)
+            self._sql_backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(
+                self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id())
+            )
         except DatabricksError as e:
             logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
             return False
@@ -371,10 +431,12 @@ class TablesMigrator:
             f"Migrating managed table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
         )
         try:
-            self._backend.execute(table_migrate_sql)
-            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+            self._sql_backend.execute(table_migrate_sql)
+            self._sql_backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(
+                self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id())
+            )
         except DatabricksError as e:
             logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
             return False
@@ -395,10 +457,12 @@ class TablesMigrator:
             table_migrate_sql = src_table.sql_migrate_ctas_external(rule.as_uc_table_key, dst_table_location)
         logger.debug(f"Migrating table {src_table.key} to {rule.as_uc_table_key} using SQL query: {table_migrate_sql}")
         try:
-            self._backend.execute(table_migrate_sql)
-            self._backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
-            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+            self._sql_backend.execute(table_migrate_sql)
+            self._sql_backend.execute(self._sql_alter_to(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(self._sql_add_migrated_comment(src_table, rule.as_uc_table_key))
+            self._sql_backend.execute(
+                self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id())
+            )
         except DatabricksError as e:
             logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
             return False
@@ -407,13 +471,15 @@ class TablesMigrator:
     def _migrate_table_in_mount(self, src_table: Table, rule: Rule):
         target_table_key = rule.as_uc_table_key
         try:
-            table_schema = self._backend.fetch(f"DESCRIBE TABLE delta.`{src_table.location}`;")
+            table_schema = self._sql_backend.fetch(f"DESCRIBE TABLE delta.`{src_table.location}`;")
             table_migrate_sql = src_table.sql_migrate_table_in_mount(target_table_key, table_schema)
             logger.info(
                 f"Migrating table in mount {src_table.location} to UC table {rule.as_uc_table_key} using SQL query: {table_migrate_sql}"
             )
-            self._backend.execute(table_migrate_sql)
-            self._backend.execute(self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id()))
+            self._sql_backend.execute(table_migrate_sql)
+            self._sql_backend.execute(
+                self._sql_alter_from(src_table, rule.as_uc_table_key, self._ws.get_workspace_id())
+            )
         except DatabricksError as e:
             logger.warning(f"failed-to-migrate: Failed to migrate table {src_table.key} to {rule.as_uc_table_key}: {e}")
             return False
@@ -429,7 +495,7 @@ class TablesMigrator:
         migrated_tables = []
         if table and not schema:
             logger.error("Cannot accept 'Table' parameter without 'Schema' parameter")
-        for cur_table in self._tc.snapshot():
+        for cur_table in self._tables_crawler.snapshot():
             if schema and cur_table.database != schema:
                 continue
             if table and cur_table.name != table:
@@ -462,8 +528,8 @@ class TablesMigrator:
             f"Reverting {table.object_type} table {table.database}.{table.name} upgraded_to {table.upgraded_to}"
         )
         try:
-            self._backend.execute(table.sql_unset_upgraded_to())
-            self._backend.execute(f"DROP {table.kind} IF EXISTS {escape_sql_identifier(target_table_key)}")
+            self._sql_backend.execute(table.sql_unset_upgraded_to())
+            self._sql_backend.execute(f"DROP {table.kind} IF EXISTS {escape_sql_identifier(target_table_key)}")
         except DatabricksError as e:
             logger.warning(f"Failed to revert table {table.key}: {e}")
 
@@ -548,7 +614,3 @@ class TablesMigrator:
             f"('upgraded_from' = '{source}'"
             f" , '{table.UPGRADED_FROM_WS_PARAM}' = '{ws_id}');"
         )
-
-    def _is_migrated(self, schema: str, table: str) -> bool:
-        index = self._migration_status_refresher.index()
-        return index.is_migrated(schema, table)
