@@ -1,7 +1,9 @@
+from __future__ import annotations
+
 import logging
 import re
 import typing
-from collections.abc import Iterable, Iterator, Collection
+from collections.abc import Collection, Iterable, Iterator
 from dataclasses import dataclass
 from enum import Enum, auto
 from functools import cached_property, partial
@@ -14,9 +16,9 @@ from sqlglot.errors import ParseError
 from databricks.labs.blueprint.parallel import Threads
 from databricks.labs.lsql.backends import SqlBackend
 from databricks.sdk.errors import NotFound
+from databricks.sdk.service.catalog import TableInfo
 
 from databricks.labs.ucx.framework.crawlers import CrawlerBase
-from databricks.labs.ucx.framework.owners import Ownership
 from databricks.labs.ucx.framework.utils import escape_sql_identifier
 
 logger = logging.getLogger(__name__)
@@ -64,6 +66,8 @@ class Table:  # pylint: disable=too-many-public-methods
     storage_properties: str | None = None
     is_partitioned: bool = False
 
+    __id_attributes__: typing.ClassVar[tuple[str, ...]] = ("catalog", "database", "name")
+
     DBFS_ROOT_PREFIXES: typing.ClassVar[list[str]] = [
         "/dbfs/",
         "dbfs:/",
@@ -85,6 +89,50 @@ class Table:  # pylint: disable=too-many-public-methods
         if isinstance(self.table_format, str):  # Should not happen according to type hint, still safer
             self.table_format = self.table_format.upper()
 
+    @classmethod
+    def from_table_info(cls, table: TableInfo) -> Table:
+        if table.catalog_name is None or table.schema_name is None or table.name is None:
+            raise ValueError(f"Catalog, schema and table name are missing: {table}")
+        if table.table_type is None:
+            raise ValueError(f"Table type is missing: {table.table_type}")
+        if table.data_source_format is None:
+            raise ValueError(f"Data source format is missing: {table.data_source_format}")
+        kwargs: dict[str, str | bool] = {
+            "catalog": table.catalog_name,
+            "database": table.schema_name,
+            "name": table.name,
+            "object_type": table.table_type.value,
+            "table_format": table.data_source_format.value,
+        }
+        if table.storage_location:
+            kwargs["location"] = table.storage_location
+        if table.view_definition:
+            kwargs["view_text"] = table.view_definition
+        if table.properties and "upgraded_to" in table.properties:
+            kwargs["upgraded_to"] = bool(table.properties.get("upgraded_to"))
+        return cls(**kwargs)  # type: ignore
+
+    @classmethod
+    def from_historical_data(cls, data: dict[str, str]) -> Table:
+        kwargs: dict[str, str | bool] = {
+            "catalog": data["catalog"],
+            "database": data["database"],
+            "name": data["name"],
+            "object_type": data["object_type"],
+            "table_format": data["table_format"],
+        }
+        if "location" in data:
+            kwargs["location"] = data["location"]
+        if "view_text" in data:
+            kwargs["view_text"] = data["view_text"]
+        if "upgraded_to" in data:
+            kwargs["upgraded_to"] = data["upgraded_to"]
+        if "storage_properties" in data:
+            kwargs["storage_properties"] = data["storage_properties"]
+        if "is_partitioned" in data:
+            kwargs["is_partitioned"] = bool(data["is_partitioned"])
+        return cls(**kwargs)  # type: ignore
+
     @property
     def is_delta(self) -> bool:
         if self.table_format is None:
@@ -362,15 +410,15 @@ class MigrationCount:
 
 
 class TablesCrawler(CrawlerBase[Table]):
-    def __init__(self, backend: SqlBackend, schema, include_databases: list[str] | None = None):
+    def __init__(self, sql_backend: SqlBackend, schema, include_databases: list[str] | None = None):
         """
         Initializes a TablesCrawler instance.
 
         Args:
-            backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
+            sql_backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
             schema: The schema name for the inventory persistence.
         """
-        super().__init__(backend, "hive_metastore", schema, "tables", Table)
+        super().__init__(sql_backend, "hive_metastore", schema, "tables", Table)
         self._include_database = include_databases
 
     def _all_databases(self) -> list[str]:
@@ -514,14 +562,14 @@ class FasterTableScanCrawler(TablesCrawler):
     Databricks workspace.
     """
 
-    def __init__(self, backend: SqlBackend, schema, include_databases: list[str] | None = None):
-        self._backend = backend
+    def __init__(self, sql_backend: SqlBackend, schema, include_databases: list[str] | None = None):
+        self._sql_backend = sql_backend
         self._include_database = include_databases
 
         # pylint: disable-next=import-error,import-outside-toplevel
         from pyspark.sql.session import SparkSession  # type: ignore[import-not-found]
 
-        super().__init__(backend, schema, include_databases)
+        super().__init__(sql_backend, schema, include_databases)
         self._spark = SparkSession.builder.getOrCreate()
 
     @cached_property
@@ -538,11 +586,18 @@ class FasterTableScanCrawler(TablesCrawler):
         return scala_option.get() if scala_option.isDefined() else None
 
     def _all_databases(self) -> list[str]:
-        try:
-            if not self._include_database:
-                return list(self._iterator(self._external_catalog.listDatabases()))
+        if self._include_database:
             return self._include_database
+        try:
+            return list(self._iterator(self._external_catalog.listDatabases()))
         except Exception as err:  # pylint: disable=broad-exception-caught
+            if "py4j.security.Py4JSecurityException" in str(err):
+                logger.error(
+                    "Failed to list databases due to Py4JSecurityException. "
+                    "Update or reinstall UCX to resolve this issue.",
+                    exc_info=True,
+                )
+                return []
             logger.error(f"failed-table-crawl: listing databases -> catalog : {err}", exc_info=True)
             return []
 
@@ -550,6 +605,13 @@ class FasterTableScanCrawler(TablesCrawler):
         try:
             return list(self._iterator(self._external_catalog.listTables(database)))
         except Exception as err:  # pylint: disable=broad-exception-caught
+            if "py4j.security.Py4JSecurityException" in str(err):
+                logger.error(
+                    "Failed to list databases due to Py4JSecurityException. "
+                    "Update or reinstall UCX to resolve this issue.",
+                    exc_info=True,
+                )
+                return []
             logger.warning(f"failed-table-crawl: listing tables from database -> {database} : {err}", exc_info=True)
             return []
 
@@ -658,13 +720,3 @@ class FasterTableScanCrawler(TablesCrawler):
         for table in table_names:
             tasks.append(partial(self._describe, catalog, database, table))
         return tasks
-
-
-class TableOwnership(Ownership[Table]):
-    """Determine ownership of tables in the inventory.
-
-    At the present we don't determine a specific owner for tables.
-    """
-
-    def _maybe_direct_owner(self, record: Table) -> None:
-        return None
