@@ -20,16 +20,11 @@ from databricks.labs.blueprint.installer import InstallState
 from databricks.labs.blueprint.parallel import ManyError, Threads
 from databricks.labs.blueprint.tui import Prompts
 from databricks.labs.blueprint.upgrades import Upgrades
-from databricks.labs.blueprint.wheels import (
-    ProductInfo,
-    Version,
-    find_project_root,
-)
+from databricks.labs.blueprint.wheels import ProductInfo, Version, find_project_root
 from databricks.labs.lsql.backends import SqlBackend, StatementExecutionBackend
 from databricks.labs.lsql.dashboards import DashboardMetadata, Dashboards
 from databricks.labs.lsql.deployment import SchemaDeployer
-from databricks.sdk import WorkspaceClient, AccountClient
-from databricks.sdk.useragent import with_extra
+from databricks.sdk import AccountClient, WorkspaceClient
 from databricks.sdk.errors import (
     AlreadyExists,
     BadRequest,
@@ -37,11 +32,11 @@ from databricks.sdk.errors import (
     InternalError,
     InvalidParameterValue,
     NotFound,
+    OperationFailed,
     PermissionDenied,
     ResourceAlreadyExists,
     ResourceDoesNotExist,
     Unauthenticated,
-    OperationFailed,
 )
 from databricks.sdk.retries import retried
 from databricks.sdk.service.dashboards import LifecycleState
@@ -51,17 +46,19 @@ from databricks.sdk.service.sql import (
     EndpointInfoWarehouseType,
     SpotInstancePolicy,
 )
+from databricks.sdk.useragent import with_extra
 
+from databricks.labs.ucx.github import DOCS_URL
 from databricks.labs.ucx.__about__ import __version__
 from databricks.labs.ucx.assessment.azure import AzureServicePrincipalInfo
 from databricks.labs.ucx.assessment.clusters import ClusterInfo, PolicyInfo
+from databricks.labs.ucx.assessment.dashboards import Dashboard
 from databricks.labs.ucx.assessment.init_scripts import GlobalInitScriptInfo
 from databricks.labs.ucx.assessment.jobs import JobInfo, SubmitRunInfo
 from databricks.labs.ucx.assessment.pipelines import PipelineInfo
 from databricks.labs.ucx.config import WorkspaceConfig
 from databricks.labs.ucx.contexts.account_cli import AccountContext
 from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext
-from databricks.labs.ucx.framework.tasks import Task
 from databricks.labs.ucx.hive_metastore.grants import Grant
 from databricks.labs.ucx.hive_metastore.locations import ExternalLocation, Mount
 from databricks.labs.ucx.hive_metastore.table_migration_status import TableMigrationStatus
@@ -78,10 +75,10 @@ from databricks.labs.ucx.runtime import Workflows
 from databricks.labs.ucx.source_code.base import UsedTable
 from databricks.labs.ucx.source_code.directfs_access import DirectFsAccess
 from databricks.labs.ucx.source_code.jobs import JobProblem
-from databricks.labs.ucx.source_code.queries import QueryProblem
+from databricks.labs.ucx.source_code.linters.queries import QueryProblem
 from databricks.labs.ucx.workspace_access.base import Permissions
 from databricks.labs.ucx.workspace_access.generic import WorkspaceObjectInfo
-from databricks.labs.ucx.workspace_access.groups import ConfigureGroups, MigratedGroup
+from databricks.labs.ucx.workspace_access.groups import AccountGroupLookup, ConfigureGroups, MigratedGroup
 
 TAG_STEP = "step"
 WAREHOUSE_PREFIX = "Unity Catalog Migration"
@@ -127,6 +124,9 @@ def deploy_schema(sql_backend: SqlBackend, inventory_schema: str):
             functools.partial(table, "directfs_in_queries", DirectFsAccess),
             functools.partial(table, "used_tables_in_paths", UsedTable),
             functools.partial(table, "used_tables_in_queries", UsedTable),
+            functools.partial(table, "inferred_grants", Grant),
+            functools.partial(table, "redash_dashboards", Dashboard),
+            functools.partial(table, "lakeview_dashboards", Dashboard),
         ],
     )
     deployer.deploy_view("grant_detail", "queries/views/grant_detail.sql")
@@ -151,7 +151,6 @@ class WorkspaceInstaller(WorkspaceContext):
         self,
         ws: WorkspaceClient,
         environ: dict[str, str] | None = None,
-        tasks: list[Task] | None = None,
     ):
         super().__init__(ws)
         if not environ:
@@ -162,7 +161,7 @@ class WorkspaceInstaller(WorkspaceContext):
             raise SystemExit(msg)
 
         self._is_account_install = self._force_install == "account"
-        self._tasks = tasks if tasks else Workflows.all().tasks()
+        self._workflows = Workflows.all()
 
     @cached_property
     def upgrades(self) -> Upgrades:
@@ -199,7 +198,7 @@ class WorkspaceInstaller(WorkspaceContext):
                 self.workspace_client,
                 self.wheels,
                 self.product_info,
-                self._tasks,
+                self._workflows,
             )
             workspace_installation = WorkspaceInstallation(
                 config,
@@ -220,7 +219,7 @@ class WorkspaceInstaller(WorkspaceContext):
             if isinstance(err.__cause__, RequestsConnectionError):
                 logger.warning(
                     f"Cannot connect with {self.workspace_client.config.host} see "
-                    f"https://github.com/databrickslabs/ucx#network-connectivity-issues for help: {err}"
+                    f"{DOCS_URL}#reference/common_challenges/#network-connectivity-issues for help: {err}"
                 )
             raise err
         return config
@@ -244,6 +243,14 @@ class WorkspaceInstaller(WorkspaceContext):
         configure_groups = ConfigureGroups(self.prompts)
         configure_groups.run()
         include_databases = self._select_databases()
+
+        skip_tacl_migration = self.prompts.confirm("Do you want to skip TACL migration when migrating tables?")
+
+        # Checking if the user wants to define a default owner group.
+        default_owner_group = None
+        if self.prompts.confirm("Do you want to define a default owner group for all tables and schemas? "):
+            default_owner_group = AccountGroupLookup(self.workspace_client).pick_owner_group(self.prompts)
+
         upload_dependencies = self.prompts.confirm(
             f"Does given workspace {self.workspace_client.get_workspace_id()} block Internet access?"
         )
@@ -251,6 +258,7 @@ class WorkspaceInstaller(WorkspaceContext):
         recon_tolerance_percent = int(
             self.prompts.question("Reconciliation threshold, in percentage", default="5", valid_number=True)
         )
+
         return WorkspaceConfig(
             inventory_database=inventory_database,
             ucx_catalog=ucx_catalog,
@@ -260,12 +268,14 @@ class WorkspaceInstaller(WorkspaceContext):
             group_match_by_external_id=configure_groups.group_match_by_external_id,  # type: ignore[arg-type]
             include_group_names=configure_groups.include_group_names,
             renamed_group_prefix=configure_groups.renamed_group_prefix,
+            skip_tacl_migration=skip_tacl_migration,
             log_level=log_level,
             num_threads=num_threads,
             include_databases=include_databases,
             trigger_job=trigger_job,
             recon_tolerance_percent=recon_tolerance_percent,
             upload_dependencies=upload_dependencies,
+            default_owner_group=default_owner_group,
         )
 
     def _compare_remote_local_versions(self):
@@ -379,7 +389,7 @@ class WorkspaceInstaller(WorkspaceContext):
         managed_table_migration_choices = {
             "Migrate MANAGED HMS table as EXTERNAL UC table. This option would require you to convert MANAGED HMS tables to EXTERNAL HMS tables once UC migration is complete, otherwise deleting HMS MANAGED table would delete the migrated UC table": 'SYNC_AS_EXTERNAL',
             "Copy data from MANAGED HMS to MANAGED UC table": 'CLONE',
-            "Convert MANAGED HMS table to EXTERNAL HMS table and migrate as EXTERNAL UC table. This risks data leakage, as once the relevant HMS tables are deleted, the underlying data won't get deleted anymore.": 'CONVERT_TO_EXTERNAL',
+            "Convert MANAGED HMS table to EXTERNAL HMS table and migrate as EXTERNAL UC table. Once the relevant HMS tables are deleted, the underlying data won't get deleted anymore, consider the impact of this change on your data workloads": 'CONVERT_TO_EXTERNAL',
         }
         managed_table_migration_choice = self.prompts.choice_from_dict(
             "If hive_metastore contains managed table with external"
@@ -492,8 +502,16 @@ class WorkspaceInstallation(InstallationMixin):
         sql_backend = StatementExecutionBackend(ws, config.warehouse_id)
         wheels = product_info.wheels(ws)
         prompts = Prompts()
-        tasks = Workflows.all().tasks()
-        workflows_installer = WorkflowsDeployment(config, installation, install_state, ws, wheels, product_info, tasks)
+        workflows = Workflows.all()
+        workflows_installer = WorkflowsDeployment(
+            config,
+            installation,
+            install_state,
+            ws,
+            wheels,
+            product_info,
+            workflows,
+        )
 
         return cls(
             config,
@@ -556,7 +574,7 @@ class WorkspaceInstallation(InstallationMixin):
             if "Unable to load AWS credentials from any provider in the chain" in str(err):
                 msg = (
                     "The UCX installation is configured to use external metastore. There is issue with the external metastore connectivity.\n"
-                    "Please check the UCX installation instruction https://github.com/databrickslabs/ucx?tab=readme-ov-file#prerequisites"
+                    f"Please check the UCX installation instruction {DOCS_URL}/installation "
                     "and re-run installation.\n"
                     "Please Follow the Below Command to uninstall and Install UCX\n"
                     "UCX Uninstall: databricks labs uninstall ucx.\n"
@@ -633,11 +651,18 @@ class WorkspaceInstallation(InstallationMixin):
     def _create_dashboard(self, folder: Path, *, parent_path: str) -> None:
         """Create a lakeview dashboard from the SQL queries in the folder"""
         logger.info(f"Creating dashboard in {folder}...")
-        metadata = DashboardMetadata.from_path(folder).replace_database(
-            database=f"hive_metastore.{self._config.inventory_database}",
-            database_to_replace="inventory",
+        metadata = (
+            DashboardMetadata.from_path(folder)
+            .replace_database(  # Assessment and migration dashboards
+                database=f"hive_metastore.{self._config.inventory_database}",
+                database_to_replace="inventory",
+            )
+            .replace_database(  # Migration progress dashboard
+                catalog=self._config.ucx_catalog,
+                catalog_to_replace="ucx_catalog",
+            )
         )
-        metadata.display_name = f"{self._name('UCX ')} {folder.parent.stem.title()} ({folder.stem.title()})"
+        metadata.display_name = self._name(f"{folder.parent.stem.title()} ({folder.stem.title()})")
         reference = f"{folder.parent.stem}_{folder.stem}".lower()
         dashboard_id = self._install_state.dashboards.get(reference)
         if dashboard_id is not None:
