@@ -1,4 +1,6 @@
 /* --title 'Total Tables' */
 SELECT
   COUNT(*) AS count_total_tables
-FROM inventory.tables
\ No newline at end of file
+FROM inventory.tables
+WHERE
+  object_type != 'VIEW'
