@@ -6,7 +6,7 @@ class MigrationRecon(Workflow):
     def __init__(self):
         super().__init__('migrate-data-reconciliation')
 
-    @job_task(job_cluster="table_migration")
+    @job_task(job_cluster="user_isolation")
     def recon_migration_result(self, ctx: RuntimeContext):
         """This workflow validate post-migration datasets against their pre-migration counterparts. This includes all
         tables, by comparing their schema, row counts and row comparison
