@@ -6,13 +6,18 @@ import itertools
 import logging
 from dataclasses import dataclass
 from pathlib import Path
-from collections.abc import Callable, Iterable, Iterator
-from typing import TypeVar, Generic
+from collections.abc import Callable, Iterable
 
 from astroid import (  # type: ignore
     NodeNG,
 )
-from databricks.labs.ucx.source_code.base import Advisory, CurrentSessionState, is_a_notebook, LineageAtom
+from databricks.labs.ucx.source_code.base import (
+    Advisory,
+    CurrentSessionState,
+    LineageAtom,
+    LocatedAdvice,
+    is_a_notebook,
+)
 from databricks.labs.ucx.source_code.python.python_ast import Tree
 from databricks.labs.ucx.source_code.path_lookup import PathLookup
 
@@ -61,17 +66,19 @@ class DependencyGraph:
     def register_import(self, name: str) -> list[DependencyProblem]:
         if not name:
             return [DependencyProblem('import-empty', 'Empty import name')]
-        maybe = self._resolver.resolve_import(self.path_lookup, name)
-        if not maybe.dependency:
-            return maybe.problems
-        maybe_graph = self.register_dependency(maybe.dependency)
+        maybe_dependency = self._resolver.resolve_import(self.path_lookup, name)
+        if maybe_dependency.problems:
+            return maybe_dependency.problems
+        assert maybe_dependency.dependency
+        maybe_graph = self.register_dependency(maybe_dependency.dependency)
         return maybe_graph.problems
 
     def register_file(self, path: Path) -> list[DependencyProblem]:
-        maybe = self._resolver.resolve_file(self.path_lookup, path)
-        if not maybe.dependency:
-            return maybe.problems
-        maybe_graph = self.register_dependency(maybe.dependency)
+        maybe_dependency = self._resolver.resolve_file(self.path_lookup, path)
+        if maybe_dependency.problems:
+            return maybe_dependency.problems
+        assert maybe_dependency.dependency
+        maybe_graph = self.register_dependency(maybe_dependency.dependency)
         return maybe_graph.problems
 
     def register_dependency(self, dependency: Dependency) -> MaybeGraph:
@@ -93,7 +100,7 @@ class DependencyGraph:
             return MaybeGraph(child_graph, [problem])
         problems = []
         for problem in container.build_dependency_graph(child_graph):
-            if problem.is_path_missing():
+            if problem.has_missing_path():
                 problem = dataclasses.replace(problem, source_path=dependency.path)
             problems.append(problem)
         return MaybeGraph(child_graph, problems)
@@ -393,8 +400,34 @@ class BaseFileResolver(abc.ABC):
 
 @dataclass
 class MaybeDependency:
+    """A class:`Dependency` or a :class:`Failure`.
+
+    The `MaybeDependency` is designed to either contain a `dependency` OR
+    `problems`, never both or neither. Typically, a `Dependency` is
+    constructed by a resolver yielding a `MaybeDependency` with
+    `list[Problems]` if the dependency could NOT be resolved, otherwise it
+    yields the `Dependency`, resulting in code that looks like:
+
+    ``` python
+    maybe_dependency = resolver.resolve_import(path_lookup, module_name)
+    if maybe_dependency.problems:
+        # Handle failure and return early
+    assert maybe_dependency.dependency, "Dependency should be given when no problems are given."
+    # Use dependency
+    ```
+    """
+
     dependency: Dependency | None
+    """The dependency"""
+
     problems: list[DependencyProblem]
+    """The problems during constructing the dependency"""
+
+    def __post_init__(self):
+        if not self.dependency and not self.problems:
+            raise ValueError(f"Dependency or problems should be given: {self}")
+        if self.dependency and self.problems:
+            raise ValueError(f"Dependency and problems should not be both given: {self}")
 
 
 class DependencyResolver:
@@ -467,7 +500,7 @@ class DependencyResolver:
     def _make_relative_paths(self, problems: list[DependencyProblem], path: Path) -> list[DependencyProblem]:
         adjusted_problems = []
         for problem in problems:
-            out_path = path if problem.is_path_missing() else problem.source_path
+            out_path = path if problem.has_missing_path() else problem.source_path
             if out_path.is_absolute() and out_path.is_relative_to(self._path_lookup.cwd):
                 out_path = out_path.relative_to(self._path_lookup.cwd)
             adjusted_problems.append(dataclasses.replace(problem, source_path=out_path))
@@ -477,25 +510,46 @@ class DependencyResolver:
         return f"<DependencyResolver {self._notebook_resolver} {self._import_resolver} {self._file_resolver}, {self._path_lookup}>"
 
 
-MISSING_SOURCE_PATH = "<MISSING_SOURCE_PATH>"
+_MISSING_SOURCE_PATH = Path("<MISSING_SOURCE_PATH>")
 
 
 @dataclass
 class DependencyProblem:
+    """A problem found with the dependency.
+
+    The line and column indexing is 0-based.
+    """
+
     code: str
+    """Unique code to identify the problem type."""
+
     message: str
-    source_path: Path = Path(MISSING_SOURCE_PATH)
-    # Lines and columns are both 0-based: the first line is line 0.
+    """A message detailing the problem."""
+
+    source_path: Path = _MISSING_SOURCE_PATH
+    """The path to the problem"""
+
     start_line: int = -1
+    """The line where the problem starts when reading source code."""
+
     start_col: int = -1
+    """The column where the problem starts when reading source code."""
+
     end_line: int = -1
+    """The line where the problem ends when reading source code."""
+
     end_col: int = -1
+    """The column where the problem ends when reading source code."""
 
-    def is_path_missing(self) -> bool:
-        return self.source_path == Path(MISSING_SOURCE_PATH)
+    def has_missing_path(self) -> bool:
+        """Flag if the path is missing, or not."""
+        return self.source_path == _MISSING_SOURCE_PATH
 
-    def as_advisory(self) -> 'Advisory':
-        return Advisory(
+    def as_located_advice(self) -> LocatedAdvice:
+        """Converts the dependency problem in a located advice for linting."""
+        # Advisory level is chosen to treat a dependency problem as a WARNING. It is the most severe while not being
+        # CRITICAL (yet).
+        advisory = Advisory(
             code=self.code,
             message=self.message,
             start_line=self.start_line,
@@ -503,6 +557,7 @@ class DependencyProblem:
             end_line=self.end_line,
             end_col=self.end_col,
         )
+        return LocatedAdvice(advisory, self.source_path)
 
     @staticmethod
     def from_node(code: str, message: str, node: NodeNG) -> DependencyProblem:
@@ -527,11 +582,16 @@ class MaybeGraph:
         return len(self.problems) > 0
 
 
+@dataclass
 class InheritedContext:
 
+    tree: Tree | None
+    found: bool
+    problems: Iterable[DependencyProblem]
+
     @classmethod
     def from_route(cls, graph: DependencyGraph, path_lookup: PathLookup, route: list[Dependency]) -> InheritedContext:
-        context = InheritedContext(None, False)
+        context = InheritedContext(None, False, [])
         for i, dependency in enumerate(route):
             if i >= len(route) - 1:
                 break
@@ -545,18 +605,6 @@ class InheritedContext:
             context = context.append(local, i == len(route) - 2)
         return context.finalize()
 
-    def __init__(self, tree: Tree | None, found: bool):
-        self._tree = tree
-        self._found = found
-
-    @property
-    def tree(self) -> Tree | None:
-        return self._tree
-
-    @property
-    def found(self) -> bool:
-        return self._found
-
     def append(self, context: InheritedContext, copy_found: bool) -> InheritedContext:
         # we should never append to a found context
         if self.found:
@@ -564,11 +612,11 @@ class InheritedContext:
         tree = context.tree
         found = copy_found and context.found
         if tree is None:
-            return InheritedContext(self._tree, found)
-        if self._tree is None:
-            self._tree = Tree.new_module()
-        self._tree.append_tree(tree)
-        return InheritedContext(self._tree, found)
+            return InheritedContext(self.tree, found, context.problems)
+        new_tree = self.tree or Tree.new_module()
+        new_tree.attach_child_tree(tree)
+        new_problems = itertools.chain(self.problems, context.problems)
+        return InheritedContext(new_tree, found, new_problems)
 
     def finalize(self) -> InheritedContext:
         # hacky stuff for aligning with Astroid's inference engine behavior
@@ -576,60 +624,17 @@ class InheritedContext:
         # see https://github.com/pylint-dev/astroid/blob/5b665e7e760a7181625a24b3635e9fec7b174d87/astroid/filter_statements.py#L113
         # this is problematic when linting code fragments that refer to inherited code with unrelated line numbers
         # here we fool the engine by pretending that all nodes from context have negative line numbers
-        if self._tree is None:
+        if self.tree is None:
             return self
-        tree = self._tree.renumber(-1)
-        return InheritedContext(tree, self.found)
+        tree = self.tree.renumber(-1)
+        return InheritedContext(tree, self.found, [])
 
 
-T = TypeVar("T")
+class StubContainer(SourceContainer):
 
+    def __init__(self, path: Path):
+        super().__init__()
+        self._path = path
 
-class DependencyGraphWalker(abc.ABC, Generic[T]):
-
-    def __init__(self, graph: DependencyGraph, walked_paths: set[Path], path_lookup: PathLookup):
-        self._graph = graph
-        self._walked_paths = walked_paths
-        self._path_lookup = path_lookup
-        self._lineage: list[Dependency] = []
-
-    def __iter__(self) -> Iterator[T]:
-        for dependency in self._graph.root_dependencies:
-            # the dependency is a root, so its path is the one to use
-            # for computing lineage and building python global context
-            root_path = dependency.path
-            yield from self._iter_one(dependency, self._graph, root_path)
-
-    def _iter_one(self, dependency: Dependency, graph: DependencyGraph, root_path: Path) -> Iterable[T]:
-        if dependency.path in self._walked_paths:
-            return
-        self._lineage.append(dependency)
-        self._walked_paths.add(dependency.path)
-        self._log_walk_one(dependency)
-        if dependency.path.is_file() or is_a_notebook(dependency.path):
-            inherited_tree = graph.root.build_inherited_tree(root_path, dependency.path)
-            path_lookup = self._path_lookup.change_directory(dependency.path.parent)
-            yield from self._process_dependency(dependency, path_lookup, inherited_tree)
-            maybe_graph = graph.locate_dependency(dependency.path)
-            # missing graph problems have already been reported while building the graph
-            if maybe_graph.graph:
-                child_graph = maybe_graph.graph
-                for child_dependency in child_graph.local_dependencies:
-                    yield from self._iter_one(child_dependency, child_graph, root_path)
-        self._lineage.pop()
-
-    def _log_walk_one(self, dependency: Dependency) -> None:
-        logger.debug(f'Analyzing dependency: {dependency}')
-
-    @abc.abstractmethod
-    def _process_dependency(
-        self,
-        dependency: Dependency,
-        path_lookup: PathLookup,
-        inherited_tree: Tree | None,
-    ) -> Iterable[T]: ...
-
-    @property
-    def lineage(self) -> list[LineageAtom]:
-        lists: list[list[LineageAtom]] = [dependency.lineage for dependency in self._lineage]
-        return list(itertools.chain(*lists))
+    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]:
+        return []
