@@ -21,7 +21,7 @@ from databricks.labs.ucx.source_code.graph import (
     DependencyGraphContext,
     InheritedContext,
 )
-from databricks.labs.ucx.source_code.python.python_analyzer import PythonCodeAnalyzer
+from databricks.labs.ucx.source_code.linters.python import PythonCodeAnalyzer
 from databricks.labs.ucx.source_code.notebooks.magic import MagicNode, MagicCommand, magic_command_factory
 
 # use a specific logger for sqlglot warnings so we can disable them selectively
@@ -84,7 +84,7 @@ class Cell(ABC):
         return f"{self.language.name}: {self._original_code[:20]}"
 
     def build_inherited_context(self, _graph: DependencyGraph, _child_path: Path) -> InheritedContext:
-        return InheritedContext(None, False)
+        return InheritedContext(None, False, [])
 
 
 class PythonCell(Cell):
@@ -437,21 +437,21 @@ class RunCommand(MagicCommand):
     def build_inherited_context(self, context: DependencyGraphContext, child_path: Path) -> InheritedContext:
         path = self.notebook_path
         if path is None:
-            logger.warning("Missing notebook path in %run command")
-            return InheritedContext(None, False)
+            problem = DependencyProblem('missing-notebook-path', "Missing notebook path in %run command")
+            return InheritedContext(None, False, [problem])
         maybe = context.resolver.resolve_notebook(context.path_lookup, path, inherit_context=True)
         if not maybe.dependency:
-            logger.warning(f"Could not load notebook {path}")
-            return InheritedContext(None, False)
+            problem = DependencyProblem('missing-notebook', f"Could not locate notebook {path}")
+            return InheritedContext(None, False, [problem])
         child_path = maybe.dependency.path
         absolute_path = context.path_lookup.resolve(path)
         absolute_child = context.path_lookup.resolve(child_path)
         if absolute_path == absolute_child:
-            return InheritedContext(None, True)
+            return InheritedContext(None, True, [])
         container = maybe.dependency.load(context.path_lookup)
         if not container:
-            logger.warning(f"Could not load notebook {path}")
-            return InheritedContext(None, False)
+            problem = DependencyProblem('corrupt-notebook', f"Could not load notebook {path}")
+            return InheritedContext(None, False, [problem])
         return container.build_inherited_context(context.parent, child_path)
 
 
