@@ -1,10 +1,10 @@
 from __future__ import annotations
 
 import builtins
-import sys
-from abc import ABC, abstractmethod
 import logging
+import sys
 import re
+from abc import ABC
 from collections.abc import Iterable
 from dataclasses import dataclass
 from typing import TypeVar, cast
@@ -16,29 +16,21 @@ from astroid import (  # type: ignore
     Call,
     ClassDef,
     Const,
-    Expr,
     Import,
     ImportFrom,
     Instance,
+    JoinedStr,
     Module,
     Name,
     NodeNG,
     parse,
     Uninferable,
-    AstroidSyntaxError,
 )
+from astroid.exceptions import AstroidSyntaxError  # type: ignore
+
+from databricks.labs.ucx.github import IssueType, construct_new_issue_url
+from databricks.labs.ucx.source_code.base import Failure
 
-from databricks.labs.ucx.source_code.base import (
-    Failure,
-    Linter,
-    Advice,
-    TableCollector,
-    UsedTable,
-    UsedTableNode,
-    DfsaCollector,
-    DirectFsAccess,
-    DirectFsAccessNode,
-)
 
 logger = logging.getLogger(__name__)
 missing_handlers: set[str] = set()
@@ -49,68 +41,87 @@ T = TypeVar("T", bound=NodeNG)
 
 @dataclass(frozen=True)
 class MaybeTree:
+    """A :class:`Tree` or a :class:`Failure`.
+
+    The `MaybeTree` is designed to either contain a `Tree` OR a `Failure`,
+    never both or neither. Typically, a `Tree` is constructed using the
+    `MaybeTree` class method(s) that yields a `Failure` if the `Tree` could
+    NOT be constructed, otherwise it yields the `Tree`, resulting in code that
+    looks like:
+
+    ``` python
+    maybe_tree = Tree.from_source_code("print(1)")
+    if maybe_tree.failure:
+        # Handle failure and return early
+    assert maybe_tree.tree, "Tree should be not-None when Failure is None."
+    # Use tree
+    ```
+    """
+
     tree: Tree | None
+    """The UCX Python abstract syntax tree object"""
+
     failure: Failure | None
+    """The failure during constructing the tree"""
 
-    def walk(self) -> Iterable[NodeNG]:
-        # mainly a helper method for unit testing
-        if self.tree is None:
-            assert self.failure is not None
-            logger.warning(self.failure.message)
-            return []
-        return self.tree.walk()
-
-    def first_statement(self) -> NodeNG | None:
-        # mainly a helper method for unit testing
-        if self.tree is None:
-            assert self.failure is not None
-            logger.warning(self.failure.message)
-            return None
-        return self.tree.first_statement()
-
-
-class Tree:
+    def __post_init__(self):
+        if self.tree is None and self.failure is None:
+            raise ValueError(f"Tree and failure should not be both `None`: {self}")
+        if self.tree is not None and self.failure is not None:
+            raise ValueError(f"Tree and failure should not be both given: {self}")
 
     @classmethod
-    def maybe_parse(cls, code: str) -> MaybeTree:
+    def from_source_code(cls, code: str) -> MaybeTree:
+        """Normalize and parse the source code to get a `Tree` or parse `Failure`."""
+        code = cls._normalize(code)
+        return cls._maybe_parse(code)
+
+    @classmethod
+    def _maybe_parse(cls, code: str) -> MaybeTree:
         try:
             root = parse(code)
             tree = Tree(root)
-            return MaybeTree(tree, None)
-        except AstroidSyntaxError as e:
-            return cls._definitely_failure('syntax-error', code, e)
-        except SystemError as e:
+            return cls(tree, None)
+        except Exception as e:  # pylint: disable=broad-exception-caught
             # see https://github.com/databrickslabs/ucx/issues/2976
-            return cls._definitely_failure('system-error', code, e)
+            failure = cls._failure_from_exception(code, e)
+            return cls(None, failure)
 
     @staticmethod
-    def _definitely_failure(message_code: str, source_code: str, e: Exception) -> MaybeTree:
-        return MaybeTree(
-            None,
-            Failure(
-                code=message_code,
-                message=f"Failed to parse code `{source_code}`: {e}. Report this as an issue on UCX GitHub.",
+    def _failure_from_exception(source_code: str, e: Exception) -> Failure:
+        if isinstance(e, AstroidSyntaxError) and isinstance(e.error, SyntaxError):
+            return Failure(
+                code="python-parse-error",
+                message=f"Failed to parse code due to invalid syntax: {source_code}",
                 # Lines and columns are both 0-based: the first line is line 0.
-                start_line=0,
-                start_col=0,
-                end_line=1,
-                end_col=1,
+                start_line=(e.error.lineno or 1) - 1,
+                start_col=(e.error.offset or 1) - 1,
+                end_line=(e.error.end_lineno or 2) - 1,
+                end_col=(e.error.end_offset or 2) - 1,
+            )
+        body = "# Desired behaviour\n\nCannot parse the follow Python code\n\n``` python\n{source_code}\n```"
+        new_issue_url = construct_new_issue_url(IssueType.BUG, "Python parse error", body)
+        return Failure(
+            code="python-parse-error",
+            message=(
+                f"Please report the following error as an issue on UCX GitHub: {new_issue_url}\n"
+                f"Caught error `{type(e)} : {e}` while parsing code: {source_code}"
             ),
+            # Lines and columns are both 0-based: the first line is line 0.
+            start_line=0,
+            start_col=0,
+            end_line=1,
+            end_col=1,
         )
 
     @classmethod
-    def maybe_normalized_parse(cls, code: str) -> MaybeTree:
-        code = cls.normalize(code)
-        return cls.maybe_parse(code)
-
-    @classmethod
-    def normalize(cls, code: str) -> str:
+    def _normalize(cls, code: str) -> str:
         code = cls._normalize_indents(code)
         code = cls._convert_magic_lines_to_magic_commands(code)
         return code
 
-    @classmethod
-    def _normalize_indents(cls, python_code: str) -> str:
+    @staticmethod
+    def _normalize_indents(python_code: str) -> str:
         lines = python_code.split("\n")
         for line in lines:
             # skip leading ws and comments
@@ -128,8 +139,8 @@ class Tree:
             return "\n".join(lines)
         return python_code
 
-    @classmethod
-    def _convert_magic_lines_to_magic_commands(cls, python_code: str) -> str:
+    @staticmethod
+    def _convert_magic_lines_to_magic_commands(python_code: str) -> str:
         lines = python_code.split("\n")
         magic_markers = {"%", "!"}
         in_multi_line_comment = False
@@ -145,10 +156,14 @@ class Tree:
                 in_multi_line_comment = not in_multi_line_comment
         return "\n".join(lines)
 
+
+class Tree:
+    """The UCX Python abstract syntax tree object"""
+
     @classmethod
     def new_module(cls) -> Tree:
         node = Module("root")
-        return Tree(node)
+        return cls(node)
 
     def __init__(self, node: NodeNG):
         self._node: NodeNG = node
@@ -167,11 +182,10 @@ class Tree:
     def walk(self) -> Iterable[NodeNG]:
         yield from self._walk(self._node)
 
-    @classmethod
-    def _walk(cls, node: NodeNG) -> Iterable[NodeNG]:
+    def _walk(self, node: NodeNG) -> Iterable[NodeNG]:
         yield node
         for child in node.get_children():
-            yield from cls._walk(child)
+            yield from self._walk(child)
 
     def locate(self, node_type: type[T], match_nodes: list[tuple[str, type]]) -> list[T]:
         visitor = MatchingVisitor(node_type, match_nodes)
@@ -191,35 +205,45 @@ class Tree:
             code = code[0:truncate_after] + "..."
         return f"<Tree: {code}>"
 
-    def append_tree(self, tree: Tree) -> Tree:
-        """returns the appended tree, not the consolidated one!"""
+    def attach_child_tree(self, tree: Tree) -> None:
+        """Attach a child tree.
+
+        1. Make parent tree of the nodes in the child tree
+        2. Extend parents globals with child globals
+
+        Attaching a child tree is a **stateful** operation for the child tree. After attaching a child
+        tree, the tree can be traversed starting from the child tree as a child knows its parent. However, the tree can
+        not be traversed from the parent tree as that node object does not contain a list with children trees.
+        """
         if not isinstance(tree.node, Module):
-            raise NotImplementedError(f"Can't append tree from {type(tree.node).__name__}")
+            raise NotImplementedError(f"Cannot attach child tree: {type(tree.node).__name__}")
         tree_module: Module = cast(Module, tree.node)
-        self.append_nodes(tree_module.body)
-        self.append_globals(tree_module.globals)
-        # the following may seem strange but it's actually ok to use the original module as tree root
-        # because each node points to the correct parent (practically, the tree is now only a list of statements)
-        return tree
+        self.attach_child_nodes(tree_module.body)
+        self.extend_globals(tree_module.globals)
 
-    def append_globals(self, globs: dict[str, list[Expr]]) -> None:
-        if not isinstance(self.node, Module):
-            raise NotImplementedError(f"Can't append globals to {type(self.node).__name__}")
-        self_module: Module = cast(Module, self.node)
-        for name, values in globs.items():
-            statements: list[Expr] = self_module.globals.get(name, None)
-            if statements is None:
-                self_module.globals[name] = list(values)  # clone the source list to avoid side-effects
-                continue
-            statements.extend(values)
+    def attach_child_nodes(self, nodes: list[NodeNG]) -> None:
+        """Attach child nodes.
 
-    def append_nodes(self, nodes: list[NodeNG]) -> None:
+        Attaching a child tree is a **stateful** operation for the child tree. After attaching a child
+        tree, the tree can be traversed starting from the child tree as a child knows its parent. However, the tree can
+        not be traversed from the parent tree as that node object does not contain a list with children trees.
+        """
         if not isinstance(self.node, Module):
-            raise NotImplementedError(f"Can't append statements to {type(self.node).__name__}")
+            raise NotImplementedError(f"Cannot attach nodes to: {type(self.node).__name__}")
         self_module: Module = cast(Module, self.node)
         for node in nodes:
             node.parent = self_module
-            self_module.body.append(node)
+
+    def extend_globals(self, globs: dict[str, list[NodeNG]]) -> None:
+        """Extend globals by extending the global values for each global key.
+
+        Extending globals is a stateful operation for this `Tree` (`self`), similarly to extending a list.
+        """
+        if not isinstance(self.node, Module):
+            raise NotImplementedError(f"Cannot extend globals to: {type(self.node).__name__}")
+        self_module: Module = cast(Module, self.node)
+        for global_key, global_values in globs.items():
+            self_module.globals[global_key] = self_module.globals.get(global_key, []) + global_values
 
     def is_instance_of(self, class_name: str) -> bool:
         for inferred in self.node.inferred():
@@ -288,6 +312,11 @@ class Tree:
         self_module: Module = cast(Module, self.node)
         return self_module.globals.get(name, None) is not None
 
+    def get_global(self, name: str) -> list[NodeNG]:
+        if not self.has_global(name):
+            return []
+        return cast(Module, self.node).globals.get(name)
+
     def nodes_between(self, first_line: int, last_line: int) -> list[NodeNG]:
         if not isinstance(self.node, Module):
             raise NotImplementedError(f"Can't extract nodes from {type(self.node).__name__}")
@@ -489,6 +518,7 @@ class MatchingVisitor(TreeVisitor):
         self._matched_nodes: list[NodeNG] = []
         self._node_type = node_type
         self._match_nodes = match_nodes
+        self._imports: dict[str, list[NodeNG]] = {}
 
     @property
     def matched_nodes(self) -> list[NodeNG]:
@@ -518,12 +548,18 @@ class MatchingVisitor(TreeVisitor):
             return
         self._matched_nodes.append(node)
 
+    def visit_joinedstr(self, node: JoinedStr) -> None:
+        if self._node_type is not JoinedStr:
+            return
+        self._matched_nodes.append(node)
+
     def _matches(self, node: NodeNG, depth: int) -> bool:
         if depth >= len(self._match_nodes):
             return False
         if isinstance(node, Call):
             return self._matches(node.func, depth)
         name, match_node = self._match_nodes[depth]
+        node = self._adjust_node_for_import_member(name, match_node, node)
         if not isinstance(node, match_node):
             return False
         next_node: NodeNG | None = None
@@ -541,6 +577,39 @@ class MatchingVisitor(TreeVisitor):
             return len(self._match_nodes) - 1 == depth
         return self._matches(next_node, depth + 1)
 
+    def _adjust_node_for_import_member(self, name: str, match_node: type, node: NodeNG) -> NodeNG:
+        if isinstance(node, match_node):
+            return node
+        # if we're looking for an attribute, it might be a global name
+        if match_node != Attribute or not isinstance(node, Name) or node.name != name:
+            return node
+        # in which case it could be an import member
+        module = Tree(Tree(node).root)
+        if not module.has_global(node.name):
+            return node
+        for import_from in module.get_global(node.name):
+            if not isinstance(import_from, ImportFrom):
+                continue
+            parent = Name(
+                name=import_from.modname,
+                lineno=import_from.lineno,
+                col_offset=import_from.col_offset,
+                end_lineno=import_from.end_lineno,
+                end_col_offset=import_from.end_col_offset,
+                parent=import_from.parent,
+            )
+            resolved = Attribute(
+                attrname=name,
+                lineno=import_from.lineno,
+                col_offset=import_from.col_offset,
+                end_lineno=import_from.end_lineno,
+                end_col_offset=import_from.end_col_offset,
+                parent=parent,
+            )
+            resolved.postinit(parent)
+            return resolved
+        return node
+
 
 class NodeBase(ABC):
 
@@ -553,131 +622,3 @@ class NodeBase(ABC):
 
     def __repr__(self):
         return f"<{self.__class__.__name__}: {repr(self._node)}>"
-
-
-class PythonLinter(Linter):
-
-    def lint(self, code: str) -> Iterable[Advice]:
-        maybe_tree = Tree.maybe_normalized_parse(code)
-        if maybe_tree.failure:
-            yield maybe_tree.failure
-            return
-        assert maybe_tree.tree is not None
-        yield from self.lint_tree(maybe_tree.tree)
-
-    @abstractmethod
-    def lint_tree(self, tree: Tree) -> Iterable[Advice]: ...
-
-
-class TablePyCollector(TableCollector, ABC):
-
-    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
-        maybe_tree = Tree.maybe_normalized_parse(source_code)
-        if maybe_tree.failure:
-            logger.warning(maybe_tree.failure.message)
-            return
-        assert maybe_tree.tree is not None
-        for table_node in self.collect_tables_from_tree(maybe_tree.tree):
-            yield table_node.table
-
-    @abstractmethod
-    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]: ...
-
-
-class DfsaPyCollector(DfsaCollector, ABC):
-
-    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
-        maybe_tree = Tree.maybe_normalized_parse(source_code)
-        if maybe_tree.failure:
-            logger.warning(maybe_tree.failure.message)
-            return
-        assert maybe_tree.tree is not None
-        for dfsa_node in self.collect_dfsas_from_tree(maybe_tree.tree):
-            yield dfsa_node.dfsa
-
-    @abstractmethod
-    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]: ...
-
-
-class PythonSequentialLinter(Linter, DfsaCollector, TableCollector):
-
-    def __init__(
-        self,
-        linters: list[PythonLinter],
-        dfsa_collectors: list[DfsaPyCollector],
-        table_collectors: list[TablePyCollector],
-    ):
-        self._linters = linters
-        self._dfsa_collectors = dfsa_collectors
-        self._table_collectors = table_collectors
-        self._tree: Tree | None = None
-
-    def lint(self, code: str) -> Iterable[Advice]:
-        maybe_tree = self._parse_and_append(code)
-        if maybe_tree.failure:
-            yield maybe_tree.failure
-            return
-        assert maybe_tree.tree is not None
-        yield from self.lint_tree(maybe_tree.tree)
-
-    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
-        for linter in self._linters:
-            yield from linter.lint_tree(tree)
-
-    def _parse_and_append(self, code: str) -> MaybeTree:
-        maybe_tree = Tree.maybe_normalized_parse(code)
-        if maybe_tree.failure:
-            return maybe_tree
-        assert maybe_tree.tree is not None
-        self.append_tree(maybe_tree.tree)
-        return maybe_tree
-
-    def append_tree(self, tree: Tree) -> None:
-        self._make_tree().append_tree(tree)
-
-    def append_nodes(self, nodes: list[NodeNG]) -> None:
-        self._make_tree().append_nodes(nodes)
-
-    def append_globals(self, globs: dict) -> None:
-        self._make_tree().append_globals(globs)
-
-    def process_child_cell(self, code: str) -> None:
-        this_tree = self._make_tree()
-        maybe_tree = Tree.maybe_normalized_parse(code)
-        if maybe_tree.failure:
-            # TODO: bubble up this error
-            logger.warning(maybe_tree.failure.message)
-            return
-        assert maybe_tree.tree is not None
-        this_tree.append_tree(maybe_tree.tree)
-
-    def collect_dfsas(self, source_code: str) -> Iterable[DirectFsAccess]:
-        maybe_tree = self._parse_and_append(source_code)
-        if maybe_tree.failure:
-            logger.warning(maybe_tree.failure.message)
-            return
-        assert maybe_tree.tree is not None
-        for dfsa_node in self.collect_dfsas_from_tree(maybe_tree.tree):
-            yield dfsa_node.dfsa
-
-    def collect_dfsas_from_tree(self, tree: Tree) -> Iterable[DirectFsAccessNode]:
-        for collector in self._dfsa_collectors:
-            yield from collector.collect_dfsas_from_tree(tree)
-
-    def collect_tables(self, source_code: str) -> Iterable[UsedTable]:
-        maybe_tree = self._parse_and_append(source_code)
-        if maybe_tree.failure:
-            logger.warning(maybe_tree.failure.message)
-            return
-        assert maybe_tree.tree is not None
-        for table_node in self.collect_tables_from_tree(maybe_tree.tree):
-            yield table_node.table
-
-    def collect_tables_from_tree(self, tree: Tree) -> Iterable[UsedTableNode]:
-        for collector in self._table_collectors:
-            yield from collector.collect_tables_from_tree(tree)
-
-    def _make_tree(self) -> Tree:
-        if self._tree is None:
-            self._tree = Tree.new_module()
-        return self._tree
