@@ -6,7 +6,6 @@ from abc import abstractmethod
 from collections.abc import Iterable, Collection
 from dataclasses import dataclass
 from datetime import timedelta
-from typing import ClassVar
 
 from databricks.labs.blueprint.limiter import rate_limited
 from databricks.labs.blueprint.parallel import ManyError, Threads
@@ -22,13 +21,15 @@ from databricks.sdk.errors.platform import (
 )
 from databricks.sdk.retries import retried
 from databricks.sdk.service import iam
-from databricks.sdk.service.iam import Group
+from databricks.sdk.service.iam import Group, User
 
 from databricks.labs.ucx.framework.crawlers import CrawlerBase
 from databricks.labs.ucx.framework.utils import escape_sql_identifier
 
 logger = logging.getLogger(__name__)
 
+SYSTEM_GROUPS: list[str] = ["users", "admins", "account users"]
+
 
 @dataclass
 class MigratedGroup:
@@ -402,7 +403,6 @@ class GroupRenameIncompleteError(RuntimeError):
 
 
 class GroupManager(CrawlerBase[MigratedGroup]):
-    _SYSTEM_GROUPS: ClassVar[list[str]] = ["users", "admins", "account users"]
 
     def __init__(  # pylint: disable=too-many-arguments
         self,
@@ -430,6 +430,7 @@ class GroupManager(CrawlerBase[MigratedGroup]):
         self._account_group_regex = account_group_regex
         self._external_id_match = external_id_match
         self._verify_timeout = verify_timeout
+        self._account_groups_lookup = AccountGroupLookup(ws)
 
     def rename_groups(self):
         account_groups_in_workspace = self._account_groups_in_workspace()
@@ -484,6 +485,9 @@ class GroupManager(CrawlerBase[MigratedGroup]):
         # Step 3: Wait for enumeration to also reflect the updated information.
         self._wait_for_renamed_groups(renamed_groups)
 
+    def current_user_in_owner_group(self, group_name: str) -> bool:
+        return self._account_groups_lookup.user_in_group(group_name, self._ws.current_user.me())
+
     def _rename_group(self, group_id: str, old_group_name: str, new_group_name: str) -> None:
         logger.debug(f"Renaming group: {old_group_name} (id={group_id}) -> {new_group_name}")
         self._rate_limited_rename_group_with_retry(group_id, new_group_name)
@@ -555,7 +559,7 @@ class GroupManager(CrawlerBase[MigratedGroup]):
 
     def reflect_account_groups_on_workspace(self):
         tasks = []
-        account_groups_in_account = self._account_groups_in_account()
+        account_groups_in_account = self._account_groups_lookup.get_mapping()
         account_groups_in_workspace = self._account_groups_in_workspace()
         workspace_groups_in_workspace = self._workspace_groups_in_workspace()
         groups_to_migrate = self.get_migration_state().groups
@@ -565,7 +569,9 @@ class GroupManager(CrawlerBase[MigratedGroup]):
                 logger.info(f"Skipping {migrated_group.name_in_account}: already in workspace")
                 continue
             if migrated_group.name_in_account in workspace_groups_in_workspace:
-                logger.error(f"Skipping {migrated_group.name_in_account}: group already exists in workspace")
+                ## After introduction of creating nested groups,
+                # we will come across groups that are already in the workspace
+                logger.warning(f"Skipping {migrated_group.name_in_account}: group already exists in workspace")
                 continue
             if migrated_group.name_in_account not in account_groups_in_account:
                 logger.warning(f"Skipping {migrated_group.name_in_account}: not in account")
@@ -626,7 +632,7 @@ class GroupManager(CrawlerBase[MigratedGroup]):
 
     def _try_fetch(self) -> Iterable[MigratedGroup]:
         state = []
-        for row in self._backend.fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
+        for row in self._sql_backend.fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
             state.append(MigratedGroup(*row))
 
         if not self._include_group_names:
@@ -646,13 +652,13 @@ class GroupManager(CrawlerBase[MigratedGroup]):
 
     def _crawl(self) -> Iterable[MigratedGroup]:
         workspace_groups_in_workspace = self._workspace_groups_in_workspace()
-        account_groups_in_account = self._account_groups_in_account()
+        account_groups_in_account = self._account_groups_lookup.get_mapping()
         strategy = self._get_strategy(workspace_groups_in_workspace, account_groups_in_account)
         yield from strategy.generate_migrated_groups()
 
     def validate_group_membership(self) -> list[dict]:
         workspace_groups_in_workspace = self._workspace_groups_in_workspace()
-        account_groups_in_account = self._account_groups_in_account()
+        account_groups_in_account = self._account_groups_lookup.get_mapping()
         strategy = self._get_strategy(workspace_groups_in_workspace, account_groups_in_account)
         migrated_groups = list(strategy.generate_migrated_groups())
         mismatch_group = []
@@ -718,17 +724,9 @@ class GroupManager(CrawlerBase[MigratedGroup]):
             groups[group.display_name] = group
         return groups
 
-    def _account_groups_in_account(self) -> dict[str, Group]:
-        groups = {}
-        for group in self._list_account_groups("id,displayName,externalId"):
-            if not group.display_name:
-                logger.debug(f"Ignoring account group in without name: {group.id}")
-                continue
-            groups[group.display_name] = group
-        return groups
-
-    def _is_group_out_of_scope(self, group: iam.Group, resource_type: str) -> bool:
-        if group.display_name in self._SYSTEM_GROUPS:
+    @staticmethod
+    def _is_group_out_of_scope(group: iam.Group, resource_type: str) -> bool:
+        if group.display_name in SYSTEM_GROUPS:
             return True
         meta = group.meta
         if not meta:
@@ -782,21 +780,6 @@ class GroupManager(CrawlerBase[MigratedGroup]):
             logger.warning(f"Group with ID {group_id} does not exist anymore in the Databricks account.")
             return None
 
-    def _list_account_groups(self, scim_attributes: str) -> list[iam.Group]:
-        # TODO: we should avoid using this method, as it's not documented
-        # get account-level groups even if they're not (yet) assigned to a workspace
-        logger.info(f"Listing account groups with {scim_attributes}...")
-        account_groups = []
-        raw = self._ws.api_client.do("GET", "/api/2.0/account/scim/v2/Groups", query={"attributes": scim_attributes})
-        for resource in raw.get("Resources", []):  # type: ignore[union-attr]
-            group = iam.Group.from_dict(resource)
-            if group.display_name in self._SYSTEM_GROUPS:
-                continue
-            account_groups.append(group)
-        logger.info(f"Found {len(account_groups)} account groups")
-        sorted_groups: list[iam.Group] = sorted(account_groups, key=lambda _: _.display_name)  # type: ignore[arg-type,return-value]
-        return sorted_groups
-
     def _delete_workspace_group_and_wait_for_deletion(self, group_id: str, display_name: str) -> str:
         logger.debug(f"Deleting workspace group: {display_name} (id={group_id})")
         self._delete_workspace_group(group_id, display_name)
@@ -929,6 +912,81 @@ class GroupManager(CrawlerBase[MigratedGroup]):
         )
 
 
+class AccountGroupLookup:
+    def __init__(self, ws: WorkspaceClient):
+        self._ws = ws
+
+    def pick_owner_group(self, prompt: Prompts) -> str | None:
+        # This method is used to select the group that will be used as the owner group.
+        # The owner group will be assigned by default to all migrated tables/schemas
+        user_id = self._ws.current_user.me().id
+        if not user_id:
+            logger.error("Couldn't find the user id of the current user.")
+            return None
+        groups = self._user_account_groups(user_id)
+        if not groups:
+            logger.warning("No account groups found for the current user.")
+            return None
+        if len(groups) == 1:
+            return groups[0].display_name
+        group_names = [group.display_name for group in groups]
+        return prompt.choice("Select the group to be used as the owner group", group_names, max_attempts=3)
+
+    def user_in_group(self, group_name: str, user: User) -> bool:
+        # This method is used to validate that the current user is a member of the group
+        user_id = user.id
+        if not user_id:
+            logger.warning("No user found for the current session.")
+            return False
+        groups = self._user_account_groups(user_id)
+        if not groups:
+            logger.warning("No account groups found for the current user.")
+            return False
+        group_names = [group.display_name for group in groups]
+        return group_name in group_names
+
+    def _user_account_groups(self, user_id: str) -> list[Group]:
+        # This method is used to find all the account groups that a user is a member of.
+        groups: list[Group] = []
+        account_groups = self._list_account_groups("id,displayName,externalId,members")
+        if not account_groups:
+            return groups
+        for group in account_groups:
+            if not group.members:
+                continue
+            for member in group.members:
+                if member.value == user_id:
+                    groups.append(group)
+                    break
+        return groups
+
+    def _list_account_groups(self, scim_attributes: str) -> list[iam.Group]:
+        # TODO: we should avoid using this method, as it's not documented
+        # get account-level groups even if they're not (yet) assigned to a workspace
+        logger.info(f"Listing account groups with {scim_attributes}...")
+        account_groups = []
+        raw = self._ws.api_client.do("GET", "/api/2.0/account/scim/v2/Groups", query={"attributes": scim_attributes})
+        for resource in raw.get("Resources", []):  # type: ignore[union-attr]
+            group = iam.Group.from_dict(resource)
+            if group.display_name in SYSTEM_GROUPS:
+                continue
+            account_groups.append(group)
+        logger.info(f"Found {len(account_groups)} account groups")
+        sorted_groups: list[iam.Group] = sorted(
+            account_groups, key=lambda _: _.display_name if _.display_name else ""
+        )  # type: ignore[arg-type,return-value]
+        return sorted_groups
+
+    def get_mapping(self) -> dict[str, Group]:
+        groups: dict[str, Group] = {}
+        for group in self._list_account_groups("id,displayName,externalId"):
+            if not group.display_name:
+                logger.debug(f"Ignoring account group in without name: {group.id}")
+                continue
+            groups[group.display_name] = group
+        return groups
+
+
 class ConfigureGroups:
     renamed_group_prefix = "db-temp-"
     workspace_group_regex = None
